import os
import glob
import numpy as np
from time import time
import sys, shutil
from numba import njit, prange

from .. import numba_plugins as nbp
from . import rotation as rot
from ...config import data_type

def setup_geometry( geo, tomogram, Omega, Kappa ):
    # define sample size and field of view
    fov = np.array( tomogram.shape[1:3], np.int32)
    nVox = np.array( tomogram.shape[:3], np.int32)

    # # convert omega/kappa rotations in lab frame to axis/angle notation via quaternions
    Q_ome = np.column_stack(
            (np.cos(Omega/2),np.outer(np.sin(Omega/2), np.array(geo.inner_axis))))
    Q_kap = np.column_stack(
            (np.cos(Kappa/2),np.outer(np.sin(Kappa/2), np.array(geo.outer_axis))))
    # zer = np.zeros_like(self.Omega)
    # Q_ome = np.column_stack(
    #         (np.cos(self.Omega/2), zer, zer, np.sin(self.Omega/2) ) )
    # #                         , sin(inner_angle)*(0,0,1)
    # Q_kap = np.column_stack(
    #         (np.cos(self.Kappa/2), zer, np.sin(self.Kappa/2), zer ) )
    Q_srot = np.array([rot.quaternion_multiply(qk,qo) for qo,qk in zip(Q_ome,Q_kap)])
    Gs = rot.OTPfromQ(Q_srot)

    # Set up translations of the beam (should make it possible to change directions no?)
    ny = fov[0]
    nz = fov[1]
    ty = np.arange(ny)-ny/2+1/2 # scan from left to right
    tz = np.arange(nz)-nz/2+1/2 # scan up to down

    # voxel coordinates
    xy = np.arange(ny)-ny/2+1/2
    z = np.arange(nz)-nz/2+1/2
    YY,XX,ZZ = np.meshgrid(xy,xy,z)
    Xb = XX.flatten()
    Yb = YY.flatten()
    Zb = ZZ.flatten()
    x_p = np.column_stack([Xb,Yb,Zb]) # voxel coordinates for (omega, kappa) = (0,0)
    return nVox, fov, Gs, ty, tz, x_p

def get_projectors(geo, Gs, sample_shape, voxel_coordinates, mask_voxels, ty, tz, shift_y, shift_z):
    """Calculates the beam intensity in each voxel for each configuration

    Attributes created
    ------------
    tomogram : 1D ndarray, float
        saves the tomogram loaded from the alignment by SASTT
        dim: 0: voxel index
    nVox : int
        number of voxels simulated
    Omega, Kappa : 1D ndarray, float
        sample rotation angles (Euler) for each unique rotation
        dim: 0: rotation index
    Gs : 2D ndarray, float
        sample rotation angles (axis-angle) for each unique rotation
        dim: 0: rotation index, 1: [omega, theta, phi]     
    Ty, Tz : 1D ndarray, float
        y- and z-distance from the central pixel for each translation
        dim: 0: translation index
    t0 : int
        translation index of the central pixel
    x_p : 2D ndarray, float
        array of coordinates of each voxel
        dim: 0: voxel index, 1: [x,y,z]
    p0 : int
        index of the central voxel (rotation center)
    mask_voxels : 1D ndarray, bool
        mask for including the voxel in the analysis or not
        dim: 0: voxel index
    cutoff_low : float
        chosen level below which voxels are masked
    Beams : 3D ndarray, float
        array of X-ray beam intensity in each voxel for each rotation
        and translation state
        dim: 0: rotation, 1: translation, 2: voxel sparse index
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel sparse index
    """

    TZ,TY = np.meshgrid(tz,ty)
    Ty,Tz = TY.flatten(), TZ.flatten()
    n_trans = Ty.shape[0]
    n_rot = Gs.shape[0]

    print('\tCalculating projectors')
    t0 = time()
    # make sparse arrays for the beams
    beam_precision = 1e-2 # cut all values below this
    Dmax = np.ceil( np.sqrt(sample_shape[0]**2+sample_shape[1]**2+sample_shape[2]**2)/2 )*2 # diagonal of the voxel-cuboid
    lb_max_approx = int(Dmax*10) # this is an estimation for the max No of entries in the sparse direction
    lb_max = 0 # will be eventually stripped to this
    Beams  = np.zeros( (n_rot, n_trans, lb_max_approx), data_type)
    Beam_idcs = np.full((n_rot, n_trans, lb_max_approx), 2**32-1, dtype=np.uint32)
    tau = np.linspace(-Dmax/2, Dmax/2, int(Dmax*2)) # parameter for the beam trajectory
    neighbors = np.transpose(np.indices((3,3,3)) - 1).reshape(-1, 3) # indices of a 3x3x3 cube
    pminmax = np.array([
        [np.min(voxel_coordinates[:,0]), np.max(voxel_coordinates[:,0])],
        [np.min(voxel_coordinates[:,1]), np.max(voxel_coordinates[:,1])],
        [np.min(voxel_coordinates[:,2]), np.max(voxel_coordinates[:,2])]
    ])

    # print('')
    t0=time()
    for g in range(n_rot): # rotations
        # rotated beam direction vector B0 = R(omega,kappa)*(1,0,0)
        B0 = np.array([1,0,0]) @ rot.MatrixfromOTP(Gs[g,0],Gs[g,1],Gs[g,2])

        # make a vector of voxels that are actually touched by the beam
        BB0, TTau = np.meshgrid( B0, tau )
        Bpath0 = np.unique(np.round(BB0*TTau),axis=0) # the closest voxels touched by the untranslated beam

        Bpath0n = np.empty( (Bpath0.shape[0]*27,3), data_type ) # the voxels above plus their neighbors
        for k in range(Bpath0.shape[0]):
            Bpath0n[k*27:(k+1)*27,:] = neighbors + Bpath0[k]
        Bpath0n = np.unique(Bpath0n, axis=0) # drop duplicates
        
        # Calculate beam intensities for every voxel and translation
        Beam_g, Beam_idcs_g = beamtranslations(
                lb_max_approx, geo.Dbeam/geo.Dstep,
                voxel_coordinates, pminmax, mask_voxels,
                B0, Bpath0n,
                Gs[g],
                Ty+shift_y[g], Tz+shift_z[g], 
                beam_precision
            )
        # Beams, iBeams = beamtranslations_full(
        #         lb_max_approx, geo.Dbeam/geo.Dstep,
        #         points_filtered,
        #         B0,
        #         Gs[g],
        #         Ty+self.shifty[g], Tz+self.shiftz[g]
        #     )
        Beam_idcs[g,:,:] = Beam_idcs_g
        Beams[g,:,:] = Beam_g

        lb = max([np.searchsorted(ib,2**32-1) for ib in Beam_idcs[g]])
        lb_max = max(lb,lb_max)
        
        t_it = (time()-t0)/(g+1)
        sys.stdout.write(f'\r\t\tProjection {(g+1):d} / {n_rot:d}, t/it: {t_it:.2f}, t left: {((n_rot-g-1)*t_it/60):.1f} min' )
        sys.stdout.flush()
    
    # strip away voxels without intensity
    Beams = Beams[:,:,:lb_max+1]
    Beam_idcs = Beam_idcs[:,:,:lb_max+1]

    print(f', finished ({(time()-t0)/60:.1f} min)' )
    return Beams, Beam_idcs

@njit(parallel=True)
def beamtranslations_full(
        beam_maxlength, dBeam,
        points,
        beam_direction,
        gs,
        Ty, Tz
    ):
    """ Calculation of beam intensities for one sample orientation
    -- inactive -- maybe this approach could speed up the calculation

    Parameters
    ----------
    beam_maxlength : int
        approximate maximum number of voxels involved in a calculation
    dBeam : float
        beam FWHM relative to the voxel size
    p : 2D ndarray, float
        dim 0: voxel index, 1: [x,y,z]
    mask_voxels : 1D ndarray, bool
        dim 0: voxel index
    B0 : 1D ndarray, float
        beam direction vector at translation 0
    gs : 1D ndarray
        sample orientation (ome,tta,phi)
    Ty, Tz : 1D ndarray, float
        beam translations with a common index
    beam_precition : float
        relative treshold for beam intensity below which values will be cut
    
    Return values
    ------------    
    values, indices
        2 arrays, together forming a sparse array

    """
    # get Gaussian standard deviation from the FWHM
    sigma = dBeam/(2*np.sqrt(2*np.log(2))) 

    # set up sparse arrays
    ntrans = Ty.shape[0]
    Beams  = np.zeros( ( ntrans, beam_maxlength), data_type)
    iBeams = (2**32-1) * np.ones( ( ntrans, beam_maxlength), np.uint32)

    for t in prange(ntrans): # translations
        # rotated translation vector (some point on the beam)
        T = np.array([0,Ty[t], Tz[t]],data_type) @ rot.MatrixfromOTP(gs[0],gs[1],gs[2])
        # get beam intensities for this configuration
        values, idcs = beamintensity(points, T, beam_direction, sigma )
        # save to sparse array
        nb = idcs.size
        Beams[t,:nb] = values
        iBeams[t,:nb] = idcs
    return Beams, iBeams

@njit(parallel=True)
def beamtranslations(
            beam_maxlength, dBeam,
            points, pminmax, mask_voxels,
            B0, Bpath0n,
            gs,
            Ty, Tz,
            beam_precision
        ):
    """ Calculation of beam intensities for one sample orientation

    Parameters
    ----------
    beam_maxlength : int
        approximate maximum number of voxels involved in a calculation
    dBeam : float
        beam FWHM relative to the voxel size
    p : 2D ndarray, float
        dim 0: voxel index, 1: [x,y,z]
    pminmax : 2D ndarray, float
        minimum and maximum voxel position for each dimension
    mask_voxels : 1D ndarray, bool
        dim 0: voxel index
    B0 : 1D ndarray, float
        beam direction vector at translation 0
    Bpath0n : 1D ndarray, int
        beam trajectory at translation 0 in voxel indices
    omega, kappa : float
        rotation angles for this projection
    Ty, Tz : 1D ndarray, float
        beam translations with a common index
    beam_precition : float
        relative treshold for beam intensity below which values will be cut
    
    Return values
    ------------    
    values, indices
        2 arrays, together forming a sparse array

    """
    ntrans = Ty.shape[0]
    Beams  = np.zeros( ( ntrans, beam_maxlength), data_type)
    iBeams = (2**32-1) * np.ones( ( ntrans, beam_maxlength), np.uint32)

    for t in prange(ntrans): # translations
        ty,tz = Ty[t], Tz[t]
        # rotated translation vector
        T = np.array([0,ty,tz],data_type) @ rot.MatrixfromOTP(gs[0],gs[1],gs[2])

        # translated beam direction vector (actually just a point on the beam)
        B = B0 + T

        # translated voxel indices supposedly touched by the beam
        Bpath = Bpath0n + T   
        # filter out voxels outside the sample
        insample = (
            (Bpath[:,0] >= pminmax[0,0] )&
            (Bpath[:,1] >= pminmax[1,0] )&
            (Bpath[:,2] >= pminmax[2,0] )&
            (Bpath[:,0] <= pminmax[0,1] )&
            (Bpath[:,1] <= pminmax[1,1] )&
            (Bpath[:,2] <= pminmax[2,1] )
        )
        Bpath = Bpath[insample]

        # voxel indices of the path points
        ib0 = (
            np.floor(Bpath[:,0]-pminmax[0,0])*(pminmax[1,1]-pminmax[1,0]+1)*(pminmax[2,1]-pminmax[2,0]+1) + \
            np.floor(Bpath[:,1]-pminmax[1,0])  *(pminmax[2,1]-pminmax[2,0]+1) + \
                np.ceil(Bpath[:,2]-pminmax[2,0]) 
            ).astype(np.uint32)                                                                        

        # filter out filled voxels by mask
        ib_msk = np.intersect1d(ib0,mask_voxels)

        # coordinates of voxels on the path
        p_path = points[ib_msk]

        # distance to beam normal to beam direction (https://mathworld.wolfram.com/Point-LineDistance3-Dimensional.html)
        w = np.empty( p_path.shape[0], data_type )
        for k in range(p_path.shape[0]):
            w[k] = np.linalg.norm(np.cross((p_path[k]-T),(p_path[k]-B)))/np.linalg.norm(B-T)

        # cumulative gaussian distribution in function of the distance from the beam
        sigma = dBeam/(2*np.sqrt(2*np.log(2))) # if dBeam is the FWHM
        beam = 1/2 * ( nbp.nb_erf( (1/2+w)/(np.sqrt(2)*sigma) ) + nbp.nb_erf( (1/2-w)/(np.sqrt(2)*sigma) ) )

        # cut away small values
        ib2 = (beam>=beam_precision)
        ib = ib_msk[ib2]
        nb = ib.shape[0] # effective number of touched voxels
        Beams[t,:nb] = beam[ib2]
        iBeams[t,:nb] = ib
    return Beams, iBeams

@njit
def beamintensity( points, beam_center, beam_direction, sigma, threshold=0.01 ):
    """Calculates a gaussian beam intensity for a given rotation and translation
    defined by the vectors beam_center (arbitrary point on the beam path) and 
    beam_direction. Values are unnormalized and cut at the value treshold.
    They are returned as sparse array, the second returned array are the indices
    which link the values to the points.

    Parameters
    ----------
    points : ndarray
        dim 0: voxel index, 1: [x,y,z]    beam_center : ndarray
    beam_direction : ndarray
        point in 3D
    sigma : float
        beam width (gaussian standard deviation)
    threshold : float, optional
        lowest value entering in the sparse array, by default 0.01

    Returns
    -------
    values, indices
        2 arrays, together forming a sparse array
    """
    beam_direction = beam_direction / np.linalg.norm(beam_direction)

    # Project each point onto the plane using the normal
    # Compute the vector from center to each point
    r = points - beam_center
    # Distance from each point to the plane
    dist_to_plane = np.dot(r, beam_direction)
    # Project onto the plane
    proj = points - np.outer(dist_to_plane, beam_direction)

    # Now compute local coordinates (u,v) in the plane to define the 2D Gaussian
    # We need two orthonormal vectors in the plane
    def orthonormal_basis(n):
        # Choose a vector not parallel to n
        if abs(n[0]) < 0.9:
            other = np.array([1, 0, 0])
        else:
            other = np.array([0, 1, 0])
        u = np.cross(n, other)
        u /= np.linalg.norm(u)
        v = np.cross(n, u)
        return u, v

    u_vec, v_vec = orthonormal_basis(beam_direction)
    # Compute coordinates in plane
    u_coords = np.dot(proj - beam_center, u_vec)
    v_coords = np.dot(proj - beam_center, v_vec)

    # Evaluate 2D Gaussian on the plane
    # gaussian = (1 / (2 * np.pi * sigma**2)) * np.exp(-(u_coords**2 + v_coords**2) / (2 * sigma**2))
    gaussian = np.exp(-(u_coords**2 + v_coords**2) / (2 * sigma**2))

    # cut away small values
    indices = np.where(gaussian > threshold)[0]
    # nb = ib.shape[0] # effective number of touched voxels
    # Beams[t,:nb] = beam[ib2]
    # iBeams[t,:nb] = ib
    return gaussian[indices], indices


# Data path and names
path_in = 'path/to/your/experiment/overview_file.h5' # .h5 file with links to the data
h5_proj_pattern = 'mysample*.1' # projection names, * is a placeholder
# .h5 internal paths:
h5_data_path = 'measurement/eiger'
## required metadata:
h5_tilt_angle_path = 'instrument/positioners/tilt' # tilt angle / None if retrieved from name (convention: .._{tilt}_{rot}_diff...h5)
h5_rot_angle_path = 'instrument/positioners/rot' # rotation angle / None if retrieved from name (... use m for minus and p for point)
h5_ty_path = 'measurement/dty' # horizontal position
h5_tz_path = 'measurement/dtz' # vertical position
### required if in scanning mode:
h5_nfast_path = 'technique/dim0' # fast axis number of points
h5_nslow_path = 'technique/dim1' # slow axis number of points
## optional metadata:
h5_ion_path = 'measurement/ion' # ionization chamber/I_0 counter if present else None
h5_data_transmission_path = 'measurement/transmission_counter' # your path or None

average_frames = 1 # sums over given frames in each dataset before integrating

# Parameters for pyFAI azimuthal integration
rad_range = [0.01, 37] # radial range
rad_unit = 'q_nm^-1' # radial parameter and unit ('q_nm^-1', ''2th_deg', etc)
azi_range = [-180, 180] # azimuthal range in degree
npt_rad = 100 # number of points radial direction
npt_azi = 120 # number of points azimuthal direction
npt_rad_1D = 2000 # number of points radial direction
int_method=('bbox','csr','opencl') # pyFAI integration methods, for GPU change 'cython' to 'opencl'
poni_path = 'path/to/your/poni_file.poni'
mask_path = 'path/to/your/mask.edf'
polarisation_factor= 0.95 # polarisation factor, usually 0.95 or 0.99
solidangle_correction = True
flatfield_correction = None #or /path/to/file
darkcurrent_correction = None #or /path/to/file

# Integration mode
mode = 2 # 1: 1D, 2: 2D, 3: both

# Parallelisation (only relevant for mode=parallel)
n_tasks = 8 # number of integrations performed in parallel
cores_per_task = 16 # size of the cluster that performs a single integration
# if opencl is used, cores_per_task is ignored


import numpy as np
import sys
import os

from .. import handle as hdl

def generators( point_group ):
    """ Provides the symmetry generators for the given point group
    by 3 angles for axis-angle parametrization [omega, theta, phi]

    Parameters
    ----------
    proper_point_group : str
        name of the crystal symmetry, taken from a provided table

    Return values
    ------------
    gen : 2D ndarray, float
        2 symmetry generators in OTP representation
        [[ome1,tta1,phi1], [ome2,tta2,phi2]]
    """
    g1 = [2*np.pi, 0, 0]; # no rotation, any mirror, glide or rotation-inversion

    g2x = [np.pi, np.pi/2, 0.]; # 2nd order rotation around [100]: _2, _2/m
    g2y = [np.pi, np.pi/2, np.pi/2]; # 2nd order rotation around [010]: 2, 2/m
    g2z = [np.pi, 0., 0.]; # 2nd order rotation around [001]: 
    g2d = [np.pi, np.pi/2, np.pi/4]; # 2nd order rotation around [110]

    g3z = [2*np.pi/3, 0., 0.]; # 3rd order rotation around [001]: 3, 3/m
    g3d = [2*np.pi/3, np.arctan(np.sqrt(2)), np.pi/4]; # 3rd order rotation around [111]: _3

    g4z = [np.pi/2, 0., 0.]; # 4th order rotation around [001]: 4, 4/m
    g4x = [np.pi/2, np.pi/2, 0.]; # 4th order rotation around [100]: 4, 4/m

    g6z = [np.pi/3, 0., 0.]; # 3rd order rotation around [001]: 6, 6/m

    match point_group: # last option is the laue group
        case ('1' | 'C1' | '-1'): # 1, m // C1
            gen = np.array((g1,g1),np.float64)
        case ('2' | 'C2' | '2/m'): # 2, 2/m, mm2, -4 // C2
            gen = np.array((g2y,g1),np.float64)
        case ('222' | 'D2' | 'mmm' ): # 222, mmm, -42m // D2
            gen = np.array((g2z,g2x),np.float64)
        case ('4' | 'C4' | '4/m'): # 4, 4/m // C4
            gen = np.array((g4z,g1),np.float64)
        case ('422' | 'D4' | '4/mmm'): # 422, 4/mmm // D4
            gen = np.array((g4z,g2x),np.float64)
        case ('3' | 'C3' | '-3' ): # 3, 3m1, 31m, -3, -6 // C3
            gen = np.array((g3z,g1),np.float64)
        case ('32' | 'D3' | '-3m' ): # 321, 312, -31m, -3m1, -6m2 // D3
            gen = np.array((g3z,g2x),np.float64)
        case ('6' | 'C6' | '6/m'): # 6, 6/m, 6mm // C6
            gen = np.array((g6z,g1),np.float64)
        case ('622' | 'D6' | '6/mmm' ): # 622, 6/mmm // D6
            gen = np.array((g6z,g2x),np.float64)
        case ('23' | 'T' | 'm3' ): # 23, -43m,  m-3 // T
            gen = np.array((g2z,g3d),np.float64)
        case ('432' | 'O' | 'm3m'): # 432, m-3m // O
            gen = np.array((g4z,g4x),np.float64)
        case _:
            print('Point group not recognized')
            sys.exit(1)
    return gen

def cSymmHSH(point_group, n):
    """Loads a matrix to transform HSHs to sHSHs

    loads sets of coefficients that make the hyperspherical 
    harmonic expansion invariant a certain crystal symmetry. 
    These can  be interpreted as defining a compact basis
    for an orientation distribution obeying the required symmetry, 
    namely, the symmetrized hyperspherical harmonics (sHSHs).

    This function loads from files created separately by Matlab
    functions in ressources/symmetrizedHSH

    Parameters
    ----------
    proper_point_group : str
        name of the crystal symmetry, taken from a provided table
    n : int
        order of the HSHs

    Return values
    ------------
    nlm : ndarray, int
        HSH indices corresponding to the coefficients
    c : 2D ndarray, complex
        matrix of HSH coefficients for all sHSHs for given n and lattice
        dimensions: 0: HSH orders, 1: sHSHs
    """
    path_csym = hdl.get_file_path('textom',
            os.path.join('ressources','symmetrizedHSH','output',
                         point_group + '_n' + str(n)))
    # filename = 'ressources/symmetrizedHSH/output/' + point_group + '_n' + str(n)
    if os.path.isfile(path_csym):
        data = np.genfromtxt(
            path_csym,
            dtype=np.complex128,
            skip_header=1,
            skip_footer=0,
        )
        
        nlm = data[:,:3].real.astype(int)
        c = data[:,3:].T
        return nlm, c
    
    else:
        print('Symmetrized HSH base file does not exist. Check for typos or generate it via matlab files')
        sys.exit(1)

def get_NsHSH(point_group, n):
    """ Returns how many sHSHs exist for the respective point group and order

    Parameters
    ----------
    point_group : str
        name of the crystal symmetry, taken from a provided table
    n : int
        order of the HSHs
    """
    if np.mod(n,2):
        return 0
    elif n==0:
        return 1
    else:
        path_overview = hdl.get_file_path('textom',
            os.path.join('ressources','symmetrizedHSH','output','overview.txt'))
        with open(path_overview, "r") as file:
                NsHSH_all = eval( file.read() )[point_group]
        return NsHSH_all[int(n/2-1)]

def get_SFnotation( point_group ):
    """Transforms proper point group from Hermann-Mauguin notation to Schonflies

    This is for connecting to oryx
    """
    convdict = {
        '1': 'C1',
        '2': 'C2',
        '222': 'D2',
        '4': 'C4',
        '422': 'D4',
        '3': 'C3',
        '32': 'D3',
        '6': 'C6',
        '622': 'D6',
        '23': 'T',
        '432': 'O',
        '-1': 'C1',
        '2/m': 'C2',
        'mmmm': 'D2',
        '4/mmm': 'D4',
        '-3m': 'D3',
        '6/mmm': 'D6',
        'm3m': 'O',

    }
    return convdict[point_group]

def get_ppg_notation( diff_laue_group ):
    """Transforms diffraction laue group to corresponding proper point group

    This is for loading the correct symmetrization matrices
    """
    convdict = {
        '-1': '1',
        '2/m': '2',
        'mmmm': '222',
        '4/mmm': '422',
        '-3m': '32',
        '6/mmm': '622',
        'm3m': '432',

    }
    try:
        return convdict[diff_laue_group]
    except:
        return diff_laue_group # this is for the case one provides a ppg

def get_proper_point_group(space_group_number):
    """
    Determine the proper point group from the space group IT number, accounting for equivalent symmetries.

    Parameters:
        space_group_number (int): The IT number of the space group.

    Returns:
        str: The proper point group symbol, or "Unknown" for invalid inputs.
    """
    # consider also:
    # from orix.quaternion.symmetry import get_point_group

    # Define ranges of space groups and their proper point groups (reduced symmetry)
    proper_point_groups = {
        "Triclinic": {range(1, 3): "1"},  # Space groups 1-2
        "Monoclinic": {range(3, 16): "2"},  # Space groups 3-15
        "Orthorhombic": {range(16, 75): "222"},  # Space groups 16-74
        "Tetragonal": {
            range(75, 89): "4",  # Space groups 75-88
            range(89, 143): "422",  # Space groups 89-142
        },
        "Trigonal": {
            range(143, 149): "3",  # Space groups 143-148
            range(149, 168): "32",  # Space groups 149-167
        },
        "Hexagonal": {
            range(168, 177): "6",  # Space groups 168-176
            range(177, 195): "622",  # Space groups 177-194
        },
        "Cubic": {
            range(195, 209): "23",  # Space groups 195-206
            range(209, 231): "432",  # Space groups 207-230
        },
    }

    # Iterate through the dictionary to find the matching range and return the proper point group
    for system, groups in proper_point_groups.items():
        for spg_range, proper_pg in groups.items():
            if space_group_number in spg_range:
                return proper_pg

    return "Unknown"  # For invalid space group numbers

def get_diffraction_laue_group(space_group_number):
    """
    Determine the proper point group from the space group IT number, accounting for equivalent symmetries.

    Parameters:
        space_group_number (int): The IT number of the space group.

    Returns:
        str: The proper point group symbol, or "Unknown" for invalid inputs.
    """
    # consider also:
    # from orix.quaternion.symmetry import get_point_group

    # Define ranges of space groups and their proper point groups (reduced symmetry)
    proper_point_groups = {
        "Triclinic": {range(1, 3): "-1"},  # Space groups 1-2
        "Monoclinic": {range(3, 16): "2/m"},  # Space groups 3-15
        "Orthorhombic": {range(16, 75): "mmm"},  # Space groups 16-74
        "Tetragonal": {
            range(75, 143): "4/mmm",  # Space groups 75-142
        },
        "Trigonal": {
            range(143, 168): "-3m",  # Space groups 143-167
        },
        "Hexagonal": {
            range(168, 195): "6/mmm",  # Space groups 168-194
        },
        "Cubic": {
            range(195, 231): "m3m",  # Space groups 195-206
        },
    }

    # Iterate through the dictionary to find the matching range and return the proper point group
    for system, groups in proper_point_groups.items():
        for spg_range, proper_pg in groups.items():
            if space_group_number in spg_range:
                return proper_pg

    return "Unknown"  # For invalid space group numbers

def zone(proper_point_group, g):
    """Computes the fundamental zone in Rodrigues space for given sample symmetries

    This function constructs a bool-array that can constrict the
    used rotations to the unique ones with respect to symmetries
    in the sample and containing crystals. The corresponding region
    in the rotation space (fundamental zone) is found by converting
    to Rodrigues parametrization and by applying the inequalities
    corresponding to the symmetry, as explained in:
    Heinz, A., and P. Neumann. Acta Crystallographica Section A 
        Foundations of Crystallography 47, no. 6 (November 1, 1991): 
        780–89. https://doi.org/10.1107/S0108767391006864.

    Parameters
    ----------
    proper_point_group : str
        name of the set of rotation symmetries, taken from a provided table
        # add numbering?
    g : 2D ndarray, float
        array of rotations given by angles: [omega, theta, phi]
        where omega is the rotation angle, theta the polar and phi
        the azimutal angle of the rotation axis
        dimensions: 0: rotation index, 1: ome,tta,phi

    Return values
    ------------
    fz : ndarray, bool
        array containing True if the corresponding rotation lies
        in the fundamental zone, False if not
    """
    # transform to Rodrigues vector components
    R1 = np.tan(g[:,0]/2) * np.sin(g[:,1]) * np.cos(g[:,2])
    R2 = np.tan(g[:,0]/2) * np.sin(g[:,1]) * np.sin(g[:,2])
    R3 = np.tan(g[:,0]/2) * np.cos(g[:,1])
    
    match proper_point_group:
        case ('1' | 'C1'): # 1, m // C1
            fz = np.full((g.shape[0]), True)

        case ('2' | 'C2'): # 2, 2/m, mm2, -4 // C2
            fz = np.logical_and.reduce((
                np.abs(R2) <= 1,
            ))
 
        case ('222' | 'D2'): # 222, mmm, -42m // D2
            fz = np.logical_and.reduce((
                np.abs(R1) <= 1,
                np.abs(R2) <= 1,
                np.abs(R3) <= 1,
            ))

        case ('4' | 'C4'): # 4, 4/m // C4
            fz = np.logical_and.reduce((
                np.abs(R1) <= 1,
                np.abs(R2) <= 1,
                +R1+R2 <= np.sqrt(2),
                -R1+R2 <= np.sqrt(2),
                -R1-R2 <= np.sqrt(2),
                +R1-R2 <= np.sqrt(2),
            ))

        case ('422' | 'D4'): # 422, 4/mmm // D4
            fz = np.logical_and.reduce((
                np.abs(R1) <= 1,
                np.abs(R2) <= 1,
                np.abs(R3) <= 1,
                R3 <= np.sqrt(2)-1,
                -R3 <= np.sqrt(2)-1,
                +R1+R2 <= np.sqrt(2),
                -R1+R2 <= np.sqrt(2),
                -R1-R2 <= np.sqrt(2),
                +R1-R2 <= np.sqrt(2),
            ))

        # tan(120deg/2) = sqrt(3)
        case ('3' | 'C3'): # 3, 3m1, 31m, -3, -6 // C3
            a = -1/2 # cos(120)
            b = np.sqrt(3)/2 # sin(120)
            fz = np.logical_and.reduce((       
                +a*R1 +b*R2 <= 1,
                -a*R1 +b*R2 <= 1,
                +a*R1 -b*R2 <= 1,
                -a*R1 -b*R2 <= 1,                
                +b*R1 +a*R2 <= 1,
                -b*R1 +a*R2 <= 1,
                +b*R1 -a*R2 <= 1,
                -b*R1 -a*R2 <= 1,        
            ))

        case ('32' | 'D3'): # 321, 312, -31m, -3m1, -6m2 // D3
            a = -1/2 # cos(120)
            b = np.sqrt(3)/2 # sin(120)
            fz = np.logical_and.reduce((      
                np.abs(R3) <= 1, # 180 deg rotation
                +a*R1 +b*R2 <= 1,
                -a*R1 +b*R2 <= 1,
                +a*R1 -b*R2 <= 1,
                -a*R1 -b*R2 <= 1,                
                +b*R1 +a*R2 <= 1,
                -b*R1 +a*R2 <= 1,
                +b*R1 -a*R2 <= 1,
                -b*R1 -a*R2 <= 1,      
            ))

        case ('6' | 'C6'): # 6, 6/m, 6mm // C6
            a = 1/2 # cos(60)
            b = np.sqrt(3)/2 # sin(60)
            fz = np.logical_and.reduce((
                np.abs(R1) <= 1, # 180 deg rotation
                np.abs(R2) <= 1, # 180 deg rotation
                +a*R1 +b*R2 <= 1,
                -a*R1 +b*R2 <= 1,
                +a*R1 -b*R2 <= 1,
                -a*R1 -b*R2 <= 1,                
                +b*R1 +a*R2 <= 1,
                -b*R1 +a*R2 <= 1,
                +b*R1 -a*R2 <= 1,
                -b*R1 -a*R2 <= 1,          
            ))

        case ('622' | 'D6'): # 622, 6/mmm // D6
            a = 1/2
            b = np.sqrt(3)/2
            fz = np.logical_and.reduce((
                np.abs(R1) <= 1, # 180 deg rotation
                np.abs(R2) <= 1, # 180 deg rotation
                np.abs(R3) <= 1, # 180 deg rotation
                +a*R1 +b*R2 <= 1,
                -a*R1 +b*R2 <= 1,
                +a*R1 -b*R2 <= 1,
                -a*R1 -b*R2 <= 1,                
                +b*R1 +a*R2 <= 1,
                -b*R1 +a*R2 <= 1,
                +b*R1 -a*R2 <= 1,
                -b*R1 -a*R2 <= 1,
                +a +b*R3 <= 1,
                -a +b*R3 <= 1,
                +a -b*R3 <= 1,
                -a -b*R3 <= 1,
                +b +a*R3 <= 1,
                -b +a*R3 <= 1,
                +b -a*R3 <= 1,
                -b -a*R3 <= 1,               
            ))

        case ('23' | 'T'): # 23, -43m,  m-3 // T
            fz = np.logical_and.reduce((
                # missing ############## ?
                R1 <= np.sqrt(2)-1,
                -R1 <= np.sqrt(2)-1,
                R2 <= np.sqrt(2)-1,
                -R2 <= np.sqrt(2)-1,
                R3 <= np.sqrt(2)-1,
                -R3 <= np.sqrt(2)-1,
            ))

        case ('432' | 'O'): # 432, m-3m // O
            fz = np.logical_and.reduce((
                R1 <= np.sqrt(2)-1,
                -R1 <= np.sqrt(2)-1,
                R2 <= np.sqrt(2)-1,
                -R2 <= np.sqrt(2)-1,
                R3 <= np.sqrt(2)-1,
                -R3 <= np.sqrt(2)-1,
                +R1+R2+R3 <= 1,
                -R1+R2+R3 <= 1,
                -R1-R2+R3 <= 1,
                -R1-R2-R3 <= 1,
                +R1-R2+R3 <= 1,
                +R1-R2-R3 <= 1,
                +R1+R2-R3 <= 1,
            ))
            
        case _:
            print('Point group not recognized')
            sys.exit(1)
    return fz

def miller_to_hex(indices):
    """
    Convert 3-index Miller indices (h,k,l) to 4-index hexagonal (h,k,i,l).
    
    Parameters
    ----------
    indices : array-like, shape (3,) or (N,3)
        Input Miller indices (h,k,l).
    
    Returns
    -------
    hex_indices : ndarray, shape (4,) or (N,4)
        Converted hexagonal indices (h,k,i,l).
    """
    arr = np.atleast_2d(indices).astype(int)
    h, k, l = arr[:, 0], arr[:, 1], arr[:, 2]
    i = -(h + k)
    hex_arr = np.stack([h, k, i, l], axis=1)
    return hex_arr if indices.ndim > 1 else hex_arr[0]


def hex_to_miller(indices, check=True):
    """
    Convert 4-index hexagonal (h,k,i,l) to 3-index Miller (h,k,l).
    
    Parameters
    ----------
    indices : array-like, shape (4,) or (N,4)
        Input hexagonal indices (h,k,i,l).
    check : bool
        If True, enforce i = -(h+k).
    
    Returns
    -------
    miller_indices : ndarray, shape (3,) or (N,3)
        Converted Miller indices (h,k,l).
    """
    arr = np.atleast_2d(indices).astype(int)
    h, k, i, l = arr[:, 0], arr[:, 1], arr[:, 2], arr[:, 3]

    if check:
        expected = -(h + k)
        if not np.all(i == expected):
            raise ValueError(f"Inconsistent indices: expected i={expected}, got {i}")

    miller_arr = np.stack([h, k, l], axis=1)
    return miller_arr if indices.ndim > 1 else miller_arr[0]


import os
import numpy as np
from time import time
import sys
import importlib as imp # for reading input files
import h5py, hdf5plugin
from numba import prange, njit
from mumott.data_handling import DataContainer
from mumott.pipelines import optical_flow_alignment as amo
from mumott.pipelines import phase_matching_alignment as amp
from mumott.pipelines import run_mitra, run_sirt
from mumott.pipelines.utilities import image_processing as imp
from scipy.ndimage import rotate, shift
from scipy.optimize import curve_fit
import multiprocessing as mp

import matplotlib.pyplot as plt

# domestic
from ..misc import import_module_from_path
from .data import get_data_list
from .. import plotting as myplt
from .. import exceptions as exc
from ..misc import cp_add_dt, timestring, find_indices_in_range
from ...config import data_type, n_threads, use_gpu

import logging
logging.getLogger("mumott").setLevel(logging.ERROR)  # Suppress some warnings from mumott

#################################################### for data import ####################################################
def mumottize( sample_dir, 
              sub_data='data_integrated', pattern='.h5', 
              align_by_transmission=False,
              q_index_range=(0,5), q_range=False,
              geo_path='input/geometry.py',
              rec_1d=False,
              ):
    
    geo = import_module_from_path('geometry', geo_path)

    if os.path.basename(os.path.dirname(sample_dir)) == 'data_integrated' :
        sample_dir = os.path.join(sample_dir,'..')

    dat, kappa, omega, airscat, dy = [], [], [], [], []

    sort_by_angles = False if geo.scan_mode == 'controt' else True
    filelist = get_data_list(sample_dir, pattern, subdir=sub_data, sort_by_angles=sort_by_angles)

    # filelist = sorted( os.listdir( os.path.join(path,sub_data)) )
    # filelist = [s for s in filelist if pattern in s]#[:2]
    # try:
    #     filelist = sort_data_list_by_angles( os.path.join(path,sub_data),
    #                                     filelist, 'tilt_angle', 'rot_angle' )
    # except:
    #     print('\t\tDid not find angles in files, just sorted data alphabetically')

    if 'controt' in geo.scan_mode:
        # continuos rotation import
        print('\tLoading data')
        all_data = []
        y_positions, z_positions = [], []
        for f, file in enumerate(filelist):
            with h5py.File(file, 'r') as hf:
                if f==0:
                    q_in = np.squeeze(hf['radial_units'][()] ).astype(data_type)
                    chi_in = np.squeeze( hf['azimuthal_units'][()]*np.pi/180 ).astype(data_type)
                    omega = ( np.array(hf['rot_angle'][()]) ).astype(data_type)
                    kappa = np.zeros_like(omega)
                    
                y_positions.append( np.mean(hf['ty'][()]) )
                z_positions.append( hf['tz'][()] )    

                d = hf['cake_integ'][()]
                if d.ndim==3:
                    all_data.append( 
                        d[:,:,q_index_range[0]:q_index_range[1]].mean(axis=1) )
                elif d.ndim==2:
                    all_data.append( 
                        d[:,q_index_range[0]:q_index_range[1]] )
                    # all_data.append( ( np.array(hf['cake_integ'][()]) ).astype(data_type) )

        # # find out y and z grids                   
        y_indices = find_unique_positions(y_positions)
        z_indices = find_unique_positions(z_positions)

        n_rot_angles = omega.size
        n_ybins = y_indices.max() + 1
        n_zbins = z_indices.max() + 1
        max_shape = (n_ybins, n_zbins)
        projections = np.zeros((n_rot_angles, n_ybins, n_zbins, *all_data[0].shape[1:])) # check size and see if this would be good for preprocess_data()
        for k,d in enumerate(all_data):
            projections[:,y_indices[k],z_indices[k],...] = d
        # and then the data is properly ordered and can go into textom
            # need to y/z check scanning directions

        print('\tCheck if data is displayed correctly, then close the figure.')
        plt.figure()
        plt.imshow(projections[0].sum(axis=2).T)
        plt.xlabel('here should be the bottom of the sample')
        plt.show(block=True)
        happy = input('\t\thappy? (y/n) ')
        if 'n' in happy:
            sys.tracebacklimit = 0
            raise exc.FOV('Data import not correct, check scanmode and consider enabling flip_fov = True')
        else:
            pass
            # return q_in

        # for k, file in enumerate( filelist ):
        #     # convert q-range to indices
        #     if np.any(q_range):
        #         with h5py.File(os.path.join(sample_dir, sub_data, file),'r') as hf:
        #             q_in = hf['radial_units'][()]
        #             q_index_range = find_indices_in_range(q_in,q_range)
        #     # continuous rotation mode
        #     with h5py.File(os.path.join(sample_dir, sub_data, file),'r') as hf:
        #         omega.append( hf['rot'][()] )
        #         kappa.append( np.zeros_like(omega) )
        #         dy.append(np.mean(hf['dty']))
        #         d = hf['cake_integ'][()]
        #     if d.ndim==3:
        #         dat.append( 
        #             d[:,q_index_range[0]:q_index_range[1]].sum(axis=2).mean(axis=1) )
        #     elif d.ndim==2:
        #         dat.append( 
        #             d[:,q_index_range[0]:q_index_range[1]].sum(axis=1) )
        #     sys.stdout.write(f'\r\t\t Projection {k+1:d} / {len(filelist):d}')
        #     sys.stdout.flush()
        # print('')
    else:
        # check if geometry is correct
        q_in = import_scan_data(sample_dir, sub_data, filelist[0],
            geo.scan_mode, q_range, q_index_range, 
            align_by_transmission=align_by_transmission, flip_fov=geo.flip_fov, rec_1d=rec_1d,
            check=True)
        # actual import
        print('\tLoading data')
        t1 = time()
        with mp.Pool(processes=n_threads) as pool:
            # Map file reading to workers
            parallel_res = pool.starmap(import_scan_data, 
                [(sample_dir, sub_data, filename, geo.scan_mode, q_range, q_index_range, align_by_transmission, geo.flip_fov,rec_1d
                  ) for filename in filelist])
        dat = [p[0] for p in parallel_res]
        kappa = np.array([p[1] for p in parallel_res], data_type)
        omega = np.array([p[2] for p in parallel_res], data_type)
        airscat = np.array([p[3] for p in parallel_res], data_type)     
        print(f'\t\ttook {(time()-t1)/60} min')

    # if 'controt' in geo.scan_mode:
    #     # sort indices according to horizontal scan
    #     scan_idx = np.argsort(dy)
    #     dat = np.array(dat)[scan_idx].T
    #     # .T[:,scan_idx]
    #     max_shape = (dat.shape[1], 1)
    #     # projections = np.zeros((dat.shape[0],*max_shape, 1), data_type)
    #     projections = dat.reshape((*dat.shape, 1, 1))

    # else:
        print('\tPadding Data')
        max_shape = (0,0)
        for d in dat:
            max_shape=(
                max(max_shape[0], d.shape[0]),
                max(max_shape[1], d.shape[1]))

        # fill projections with air scattering
        projections = np.array(
            [ np.random.normal(a,np.sqrt(a)/2,(*max_shape, dat[0].shape[2])) for a in airscat], 
            data_type)
        # put data padded with the airscattering
        for k in range(len(dat)):
            # med_dat = np.median(dat[k])
            # dat[k][dat[k] > 10000*med_dat] = med_dat # filter crazy pixels
            si, sj = dat[k].shape[:2]
            i0 = (max_shape[0] - si)//2
            j0 = (max_shape[1] - sj)//2
            projections[k, i0:i0+si, j0:j0+sj] = dat[k]

    print('\tRemoving outliers')
    projections = _remove_crazypixels( projections )

    if rec_1d: 
        # make a h5 file containing all the projections for all q-values
        mm_datafile = os.path.join(sample_dir,'analysis','rec_1d','data_rec1d.h5')
        with h5py.File(mm_datafile,'w') as hf:
            hf.create_dataset( 'radial_units', data= np.array(q_in) )
            hf.create_dataset( 'projections', 
                              (projections.shape[3], *projections.shape[:3]),
                              chunks = (1, *projections.shape[:3]),
                              shuffle="True", compression="lzf",
                              data=np.transpose(projections, (3,0,1,2) )
                            )

    else:
        # make a mumott h5 file for alignment
        # pretending that the projections are the inverse of an absorbtion measurement
        diod = projections.mean(axis=3)
        diod = 1-diod/diod.max()
        diod[diod > 1] = 1.

        os.makedirs( os.path.join(sample_dir,'analysis/'), exist_ok=True )
        mm_datafile = os.path.join(sample_dir,f'analysis/data_mumott.h5')

        print(f'\tSaving mumott data file to {mm_datafile}')
        # t0=time()
        with h5py.File(mm_datafile,'w') as hf:
            hf.create_dataset( 'detector_angles', data= np.array([0]) ) 
            hf.create_dataset( 'detector_direction_origin', data=np.array(
                geo.detector_direction_origin, data_type) )
            hf.create_dataset( 'detector_direction_positive_90', data=np.array(
                geo.detector_direction_positive_90, data_type) )
            hf.create_dataset( 'inner_axis', data=np.array(
                geo.inner_axis, data_type) )
            hf.create_dataset( 'outer_axis', data=np.array(
                geo.outer_axis, data_type) )
            hf.create_dataset( 'j_direction_0', data=np.array(
                geo.transverse_horizontal, data_type) )
            hf.create_dataset( 'k_direction_0', data=np.array(
                geo.transverse_vertical, data_type) )
            hf.create_dataset( 'p_direction_0', data=np.array(
                geo.beam_direction, data_type) )
            hf.create_dataset( 'volume_shape', data=np.array((max_shape[0],max_shape[0],max_shape[1]), np.int32) )
            hf.create_dataset( 'fov_flipped', data=geo.flip_fov)

            gp = hf.create_group('projections')
            for k, proj in enumerate(projections):
                gpk = gp.create_group(str(k))
                gpk.create_dataset('data', data=proj)#.reshape( proj.shape ))
                gpk.create_dataset('diode', data=diod[k])
                gpk.create_dataset('inner_angle', data=omega[k]*np.pi/180)
                gpk.create_dataset('outer_angle', data=kappa[k]*np.pi/180)
                gpk.create_dataset('j_offset', data=0.)
                gpk.create_dataset('k_offset', data=0.)
        # print(f'\t\ttook {time()-t0:.1f} s')
    return mm_datafile

def get_q_index_range(path, sub_data, file,q_range):
    # convert q-range to indices
    if np.any(q_range):
        with h5py.File(os.path.join(path, sub_data, file),'r') as hf:
            q_in = hf['radial_units'][()].flatten()
            q_index_range = find_indices_in_range(q_in,q_range)
    return q_index_range

def find_unique_positions( positions ):
    """This assumes that positions are on a 1d grid but there is noise.

    Finds out the grid points and puts them in ascending order.
    Categorizes the positions into the grid points and returns the indices in the grid.

    Parameters
    ----------
    positions : ndarray, float

    Returns
    ----------
    indices: 1darray, int
        corresponding indices for each position to the grid
    """
    positions = np.array(positions).flatten()
    diffs = np.diff(np.sort(positions))
    if diffs.max() > 5*diffs.min(): # if there is several points at the same grid point
        threshold = np.median(diffs) + np.std(diffs)
        step = np.median(diffs[diffs > threshold])  # ignore steps that come from noise
        a0 = np.min(positions)
        # Quantize each value to its nearest grid index
        indices = np.round((np.array(positions) - a0) / step).astype(int)
        # unique_idx = np.unique(idx)
    else:
        # only one z-position
        indices = np.argsort(positions) 
    return indices

def import_scan_data(
        sample_dir, sub_data, file,
        scan_mode,
        q_range, q_index_range,
        align_by_transmission=False,
        flip_fov=False,rec_1d=False,
        check=False
    ):

    if np.any(q_range):
        q_index_range = get_q_index_range(sample_dir, sub_data, file, q_range)

    with h5py.File(os.path.join(sample_dir, sub_data, file),'r') as hf:
        q_in=( hf['radial_units'][0,slice(*q_index_range)] 
                ).astype(data_type).flatten()
        fov = ( hf['fov'][()] ).astype(np.int32)
        try:
            kappa = ( hf['tilt_angle'][()] )
            omega = ( hf['rot_angle'][()] )

        except: # for backwards compatibilty
            # find out angles from name:
            splnm = file.split('_')
            i_diff = next(i for i, s in enumerate(splnm) if 'diff' in s) # find where is 'diff'
            kappa = float( splnm[i_diff-2].replace('m','-') ) 
            omega = float( splnm[i_diff-1].replace('p','.') ) 
        # d = hf['cake_integ'][:,:,q_index_range[0]:q_index_range[1]].sum(axis=2)
        # d = d.reshape(*fov,d.shape[1])
        if align_by_transmission:
            d = np.atleast_2d(hf['data_transmission'][()].astype(data_type)).T
        else:
            d = hf['cake_integ']
            if d.ndim==2: # for reading 1D data
                d = d[:,q_index_range[0]:q_index_range[1]].astype(data_type)
            elif d.ndim==3:
                d = d[:,:,q_index_range[0]:q_index_range[1]].astype(data_type).mean(axis=1)

        # automatically recognize hor. and vert. scan direction
        ty = np.atleast_1d(hf['ty'][()])
        y_scan_direction = np.sign(ty[-1] - ty[0])
        tz = np.atleast_1d(hf['tz'][()])
        z_scan_direction = np.sign(tz[-1] - tz[0])

        idx_data = np.arange(d.shape[0])

    if flip_fov:
        fov = np.flip(fov)

    if rec_1d: # for q-resolved data
        d = d.reshape(*fov,d.shape[1])
    else:
        d = d.sum(axis=1)
        d = d.reshape(*fov,1)
    idx_data = idx_data.reshape(fov)

    if 'line' in scan_mode:
        fov = np.flip( fov )
        # d = np.fliplr(np.flipud( np.transpose( d, axes=(1,0,2)) ) )
        d = np.transpose( d, axes=(1,0,2))
        # d = np.rot90(d, 2)
        idx_data = np.transpose( idx_data )

    # reorder data so that they are all treated the same in the model
    if 'snake' in scan_mode:
        # Flip every second row
        for ii in range(d.shape[0]):
            if ii % 2 != 0:
                d[ii] = d[ii][::-1]
                idx_data[ii] = idx_data[ii][::-1]

    # account for hor. and vert. scan direction
    if y_scan_direction < 0:
        d = np.flip(d, axis=0)
        idx_data = np.flip(idx_data, axis=0)
    if z_scan_direction < 0:
        idx_data = np.flip(idx_data, axis=1)

    # get airscattering:
    airscat = min( d[0,0].mean(), d[0,-1].mean(), d[-1,0].mean(), d[-1,-1].mean() )

    if check:
        print('\tCheck if data is displayed correctly, then close the figure.')
        plt.figure()
        plt.imshow(d.mean(-1).T)
        plt.xlabel('here should be the bottom of the sample')
        plt.show(block=True)
        happy = input('\t\thappy? (y/n) ')
        if 'n' in happy:
            sys.tracebacklimit = 0
            raise exc.FOV('Data import not correct, check scanmode and consider enabling flip_fov = True')
        else:
            return q_in
    else:
        return d, kappa, omega, airscat, idx_data

#################################################### for actual alignment ####################################################
def align( path_in, 
        mode = 'optical_flow',
        rec_iteration=5, stop_max_iteration=20, 
        pm_upsampling=10, excluded_projections=[],
        offsets_start = None, center_rec=True,
        redo=False, use_gpu=True,
        align_horizontal=True, align_vertical=True,
        crop_image=False, do_align=True,
        output = True, ):

    dc = DataContainer( path_in )
    d = dc.data[:, :, :, 0]
    diodes_ref = np.moveaxis( d, 0, -1)
    if do_align:
        if np.any(offsets_start):
            # take given offsets as start values
            dc.geometry.j_offsets = np.array(offsets_start[0], dtype=data_type)
            dc.geometry.k_offsets = np.array(offsets_start[1], dtype=data_type)
        if redo:
            # remove all prior alignment from geometry
            dc.geometry.j_offsets = np.zeros(len(dc.geometry), dtype=data_type)
            dc.geometry.k_offsets = np.zeros(len(dc.geometry), dtype=data_type)

        alignment_param = dict(
                use_gpu = use_gpu,
                # optimal_shift=np.zeros((np.size(dc.diode, 0), 2)),
                rec_iteration=rec_iteration,
                stop_max_iteration=stop_max_iteration,
                align_horizontal=align_horizontal, align_vertical=align_vertical,
                center_reconstruction=center_rec,
                # high_pass_filter=0.01,    
                # smooth_data = True,
                # sigma_smooth = 3,
                optimizer_kwargs=dict(nestorov_weight=0.6)
            )

        if mode == 'optical_flow':
            shifts, sinogram, tomogram = amo.run_optical_flow_alignment( dc, **alignment_param )
            shifts += np.column_stack( (dc.geometry.j_offsets,dc.geometry.k_offsets) )

        elif mode == 'phase_matching':
            if np.any(crop_image):
                projection_cropping = (slice(*crop_image[0]), slice(*crop_image[1]))
            else:
                projection_cropping = (slice(None,None,None),slice(None,None,None))
            result = amp.run_phase_matching_alignment(dc, 
                    ignored_subset=None, 
                    projection_cropping=projection_cropping,
                    # projection_cropping=(slice(None, None, None), slice(None, None, None)), 
                    # reconstruction_pipeline=<function run_mitra>, 
                    # reconstruction_pipeline_kwargs=None, 
                    use_gpu=True, use_absorbances=True, 
                    maxiter=stop_max_iteration, upsampling=pm_upsampling, 
                    shift_tolerance=None, 
                    shift_cutoff=None, relative_sample_size=1.0, 
                    relaxation_weight=0.0, center_of_mass_shift_weight=0.0, 
                    align_j=True, align_k=True
                    )
            shifts = np.column_stack( (dc.geometry.j_offsets,dc.geometry.k_offsets) )
            sinogram = np.moveaxis(result['projections'][:,:,:,0], 0, -1)
            tomogram = result['reconstruction']

        # shape j x k x p
        diodes_shifted = imp.imshift_fft(diodes_ref, shifts[:, 0], shifts[:, 1])

    else:
        shifts = np.zeros( (len(dc.geometry.j_offsets), 2), data_type )
        fov = dc.projections[0].diode.shape
        tomogram = np.ones( (fov[0],fov[0],fov[1]) )
        sinogram=0
        mode='no_alignment'

    if output:
        path_out = os.path.join( os.path.dirname(path_in), 'alignment_result.h5' )
        with h5py.File( path_out, 'w') as hf:
            hf.create_dataset('shifts', data=shifts)
            hf.create_dataset('sinogram', data=sinogram)
            hf.create_dataset('tomogram', data=tomogram )
            if do_align:
                hf.create_dataset('unshifted_data', data=diodes_ref )
                hf.create_dataset('shifted_data', data=diodes_shifted )
            hf.create_dataset('omega', data=dc.geometry.inner_angles )
            hf.create_dataset('kappa', data=dc.geometry.outer_angles )
            hf.create_dataset('alignment_mode', data=mode)
            if np.any(crop_image):
                hf.create_dataset('cropping', data=crop_image)
            
        print(f'Saved alignment result to {path_out}')
    
    return shifts

def regroup( h5path, regroup_max = 16, 
            horizontal=True, vertical=True 
            ):
    print('\tDownsample data')
    path_dir = os.path.dirname(h5path)

    att_cp = [ 'detector_direction_origin', 'detector_direction_positive_90', 
              'inner_axis', 'outer_axis', 'j_direction_0', 'k_direction_0', 'p_direction_0',]

    for k in range( 1, int(np.log2(regroup_max))+1 ):
        print(f'\t\t{2**k:d} times')
        new_h5path = os.path.join(path_dir,f'data_mumott_gr{2**k:d}.h5')

        with h5py.File(h5path, 'r') as src, h5py.File(new_h5path, 'w') as dst:
            # copy parameters from the other h5 file
            def copy_items(name, obj):
                if name in att_cp:
                    src.copy(name, dst)
            src.visititems(copy_items)
            
            # regroup data in every projection by taking means of 2x2 blocks
            proj = src['projections']
            projNo = []
            def collect_projNo(name, obj):
                if isinstance(obj, h5py.Group):
                    projNo.append(name)
            proj.visititems(collect_projNo)
            projNo = sorted(projNo, key=lambda x: int(x))
            gp = dst.create_group('projections')
            for k, pN in enumerate(projNo):
                diode_new = _average_neighbors(src[f'projections/{pN}/diode'][()],
                                              horizontal,vertical)
                data_new = (1-diode_new).reshape((*diode_new.shape,1))
                gpk = gp.create_group(str(k))
                gpk.create_dataset('data', data=data_new)
                gpk.create_dataset('diode', data=diode_new)
                gpk.create_dataset('inner_angle', data=src[f'projections/{pN}/inner_angle'])
                gpk.create_dataset('outer_angle', data=src[f'projections/{pN}/outer_angle'])
                gpk.create_dataset('j_offset', data=0.)
                gpk.create_dataset('k_offset', data=0.)

            dst.create_dataset('detector_angles', data=np.array([0.]), dtype=data_type)
            dst.create_dataset('volume_shape', dtype=np.int32,
                               data=[diode_new.shape[0],diode_new.shape[0],diode_new.shape[1]])
        h5path = new_h5path # create the next one from the last

def align_regrouped( path:str, mode='optical_flow',
                    regroup_max = 16, use_gpu=True,
                    pre_rec_it = 5, pre_max_it = 5,
                    last_rec_it = 40, last_max_it = 5,
                    align_horizontal=True, align_vertical=True,
                    crop_image = False, do_align=True
                     ):
    """Aligns dataset step by step going to higher and higher sampling

    Parameters
    ----------
    path : str
        sample_dir/analysis/
    mode : str, optional
        _description_, by default 'optical_flow'
    regroup_max : int, optional
        _description_, by default 16
    use_gpu : bool, optional
        _description_, by default True
    pre_rec_it : int, optional
        _description_, by default 5
    pre_max_it : int, optional
        _description_, by default 5
    last_rec_it : int, optional
        _description_, by default 40
    last_max_it : int, optional
        _description_, by default 5
    align_horizontal : bool, optional
        _description_, by default True
    align_vertical : bool, optional
        _description_, by default True
    crop_image : bool, optional
        _description_, by default False
    """
    # check how many projections there are
    subgroup_count = 0
    def collect_subgroups(name, obj):
        nonlocal subgroup_count
        if isinstance(obj, h5py.Group) and name != 'projections':
            subgroup_count += 1
    with h5py.File(os.path.join(path,f'data_mumott.h5'), 'r') as f:
        group = f['projections']
        group.visititems(collect_subgroups)

    # load excluded projection indices
    # excl = np.genfromtxt(os.path.join(path,))
    
    offsets = np.zeros((2,subgroup_count),data_type)
    all_shifts = []

    if regroup_max > 1 and do_align:
        for k in range( int(np.log2(regroup_max)), 0, -1 ):
            regroup_pxls = 2**k
            if crop_image:
                cr=(np.array(crop_image)/regroup_pxls).astype(np.int32)
                cr[np.array(crop_image)==-1]=-1
            else:
                cr = False
            print(f'Align {regroup_pxls:d}-times downsampled data')
            shifts = align( os.path.join(path,f'data_mumott_gr{regroup_pxls:d}.h5'),
                    mode=mode, crop_image=cr, #excluded_projections=excl,
                    rec_iteration=pre_rec_it, stop_max_iteration=pre_max_it, 
                    align_horizontal=align_horizontal, align_vertical=align_vertical,
                    use_gpu=use_gpu,offsets_start=offsets,
                    output=False, redo=False )
            
            offsets = (2*shifts).T
            all_shifts.append(offsets * 2**(k-1))

    if do_align:
        print('Align full dataset')
    shifts = align( os.path.join(path,f'data_mumott.h5'),
                    mode=mode, crop_image=crop_image,
                    rec_iteration=last_rec_it,
                    stop_max_iteration=last_max_it, use_gpu=use_gpu,
                    center_rec=False, do_align=do_align,
                    offsets_start=offsets , output=True, redo=False)
    all_shifts.append(shifts.T)

    if do_align:
        fig,ax = plt.subplots(2,1, figsize=(8,4), sharex=True)
        for k,a in enumerate(all_shifts):
            ax[0].plot(a[0],label=str(2**k))
            ax[1].plot(a[1])
        ax[0].legend(title='downsampling')
        ax[0].set_ylabel('Horizontal shift [pixels]')
        ax[1].set_ylabel('Vertical shift [pixels]')
        ax[1].set_xlabel('Projections')
        fig.tight_layout()
        plt.show()

def project_tomogram(tomogram, rotation_angle, tilt_angle, axis=(1, 2)):
    """
    Project a 3D tomogram given a rotation and tilt angle.

    Parameters:
    tomogram (ndarray): 3D numpy array representing the tomogram.
    rotation_angle (float): Rotation angle in rad.
    tilt_angle (float): Tilt angle in rad.
    axis (tuple): Axis around which to rotate the tomogram. Default is (1, 2).

    Returns:
    projection (ndarray): 2D numpy array representing the projection.
    """
    # Step 1: Rotate the tomogram
    rotated_tomogram = rotate(tomogram, rotation_angle*180/np.pi, 
                              axes=axis, reshape=False)

    # Step 2: Tilt the tomogram
    tilted_tomogram = rotate(rotated_tomogram, tilt_angle*180/np.pi, 
                             axes=(0, 2), reshape=False)

    # Step 3: Project the tomogram onto a 2D plane
    projection = np.sum(tilted_tomogram, axis=0)

    return projection

def get_projections( g, path_mumott_data, path_alignment ):
    # Load the actual data
    with h5py.File( path_mumott_data, 'r') as hf:
        proj = hf['projections'][str(g)]['diode'][()]
    proj = 1 - proj

    #### This is to double-check the projection with scipy, needs the additional inputs:
    ## tomogram, kap, ome, shift_y, shift_z 
    ## e.g. from this:     kap,ome,shift_y,shift_z,tomogram = hdl.load_shifts_mumott( path_alignment )
    # # Step 1: Rotate the tomogram
    # rotated_tomogram = rotate(tomogram, ome[g]*180/np.pi, axes=(0,1), reshape=False)

    # # Step 2: Tilt the tomogram
    # tilted_tomogram = rotate(rotated_tomogram, -kap[g]*180/np.pi, axes=(0, 2), reshape=False)

    # # Step 3: Project the tomogram onto a 2D plane
    # proj_n = np.sum(tilted_tomogram, axis=0)

    # # Shift the projection for comparison
    # proj_n = proj_n.reshape(proj_n.shape[:2])# * proj.max()/proj_n.max()
    # proj_n = shift( proj_n, shift=(-shift_y[g], -shift_z[g]), 
    #             order=1, mode='nearest')

    # Load the projection
    with h5py.File( path_alignment, 'r') as hf:
        proj_n = hf['sinogram'][:,:,g]
    
    # rescale projection
    def linear_func(x, a):
        return a * x
    popt, _ = curve_fit(linear_func, proj.flatten(), proj_n.flatten())
    proj_n /= popt[0]

    return proj, proj_n

def check_consistency( path_alignment, path_MMdata, save_to_result=False ):
    print('\tCalculating residuals')
    err = []
    with h5py.File(path_alignment,'r') as hf:
        kap = hf['kappa'][()]
    for g in range( kap.size ):
        proj,proj_n = get_projections( g, path_MMdata, path_alignment )
        err.append( np.sqrt(((proj-proj_n)**2).sum()/(proj**2).sum()) )
        sys.stdout.write('\r\t\t %d / %d' % (g+1,kap.size))
        sys.stdout.flush()

    if save_to_result:
        with h5py.File(path_alignment,'r+') as hf:
            hf.create_dataset( 'consistency', data=err )

    print('')
    plt.figure()
    plt.plot(err)
    plt.xlabel( 'Projection' )
    plt.ylabel( 'Average relative deviation' )
    plt.title( 'Consistency of measured and simulated projections' )
    plt.show()

def show_aligned_projection( g, path_alignment, path_MMdata ):
    with h5py.File(path_alignment,'r') as hf:
        kap = hf['kappa'][()]
        ome = hf['omega'][()]
    proj,proj_n = proj,proj_n = get_projections( g, path_MMdata, path_alignment )
    fig,ax = plt.subplots(1,3, figsize=(9,4))
    vmax = max( proj.max(), proj_n.max() )
    vmin = min( proj.min(), proj_n.min() )
    ax[0].imshow(proj_n, cmap='jet', vmin=vmin, vmax=vmax )
    ax[1].imshow(proj, cmap='jet', vmin=vmin, vmax=vmax)
    ax[2].imshow(np.abs(proj_n-proj), cmap='jet', vmin=vmin, vmax=vmax)
    ax[0].set_title('Projected from tomogram')
    ax[1].set_title('Measured')
    ax[2].set_title('Absolute deviation')
    fig.suptitle(f'Projection {g}: kap {kap[g]*180/np.pi:.1f}, ome: {ome[g]*180/np.pi:.1f}')
    fig.tight_layout()
    plt.show()

def check_projections( path, 
                      path_result = 'analysis/alignment_result.h5',
                      path_data = 'analysis/data_mumott.h5',
                      output = True,
                       ):

    print('\tLoading projections and shifts')
    with h5py.File(os.path.join(path,path_result),'r') as hf:
        shifts = hf['shifts'][()]

    with h5py.File(os.path.join(path,path_data), 'r') as hf:
        gr = hf['projections']
        projNo = []
        def collect_projNo(name, obj):
            if isinstance(obj, h5py.Group):
                projNo.append(name)
        gr.visititems(collect_projNo)
        projNo = sorted(projNo, key=lambda x: int(x))
        proj, ome, kap = [], [], []
        for pN in projNo:
            proj.append( gr[pN]['diode'][()] )
            ome.append( gr[pN]['inner_angle'][()])
            kap.append(gr[pN]['outer_angle'][()])

    print('\tPlotting shifted projections in browser (check popup blocker)')
    myplt.plot_shifted_projections(
        1-np.array(proj), shifts[:,1], shifts[:,0],
        names= [f'omega {180/np.pi*o:.1f}, kappa {180/np.pi*k:.1f}' for o,k in zip(ome,kap)],
        output=output,
    )

def backproject_all( sample_dir, path_input, range_q=None, path_data=None, only_mumottize=False, batch_size=10 ):
    """calculates a scalar tomogram for each q-value

    Parameters
    ----------
    path : str  
        textom base sample directory    
    path_input : str    
        path of the input file
    range_q : list or ndarray or None
        first and second entry will be evaluated as indices
    path_data : str, optional
        path to the data_mumott.h5 file, if None makes it, by default None
    only_mumottize : bool, optional
        only makes the data_mumott.h5 file, by default False
    batch_size : int, optional
        decides how many datasets are loaded at once (memory optimization), by default 10

    Returns
    -------
    str
        path of the result .h5 file
    """
    inp = import_module_from_path('input_1Drec', path_input )
    
    with h5py.File( os.path.join(sample_dir,'analysis/alignment_result.h5'), 'r' ) as hf:
        # kap = hf['kappa'][()]
        # ome = hf['omega'][()]
        shifts = hf['shifts'][()]

    # kap=np.delete( kap, inp.exclude_proj )
    # ome=np.delete( ome, inp.exclude_proj )
    # shifts=np.delete( shifts, inp.exclude_proj, axis=0 )
 
    if not path_data:
        path_data = mumottize(sample_dir, sub_data=inp.subdir_data, pattern=inp.pattern, 
            q_index_range=[0,None],
            geo_path=os.path.join(sample_dir,'analysis/geometry.py'),
            rec_1d=True,
            )
    
    if not only_mumottize:
        # load mumott file from alignment
        dc = DataContainer(os.path.join(sample_dir,'analysis/data_mumott.h5'))

        #Load metadata
        with h5py.File(path_data, 'r') as hf:
            nq, n_proj, fov_y, fov_z = hf['projections'].shape
            q_values = hf['radial_units'][()]
        
        if range_q:
            q_values = q_values[int(range_q[0]):int(range_q[1])]
            nq = q_values.size
        else:
            range_q = [0,-1]

        # set up output file
        title = os.path.basename(os.path.dirname(sample_dir))
        out_path = os.path.join(sample_dir,'analysis/rec_1d/',
                        f'reconstruction_1Dmm_{title}_{timestring()}.h5')
        with h5py.File( out_path, 'w' ) as hf:
            hf.create_dataset( 'sample_name', data=title )
            hf.create_dataset( 'radial_units', data=q_values )
            hf.create_dataset( 'sinogram', 
                              shape=(nq, n_proj, fov_y, fov_z), dtype=data_type,
                              compression=hdf5plugin.LZ4()
                                )
            hf.create_dataset( 'tomogram', 
                              shape=(nq, fov_y, fov_y, fov_z), dtype=data_type,
                               compression=hdf5plugin.LZ4() )

        n_batches = nq//batch_size
        for l in range( n_batches ):
            print('\tLoading batch')
            with h5py.File(path_data, 'r') as hf:
                projections = hf['projections'][int(range_q[0])+l*batch_size:int(range_q[0])+(l+1)*batch_size]
            sino, tomo = [], []
            # reconstruct each q-value
            print('\tStarting reconstructions')
            for k in range( batch_size ):
                projection = projections[k]
                # transform to "diode scale"
                scal = projection.max()
                d = 1 - projection/scal

                # insert the data shifted according to alignment
                diodes_ref = np.moveaxis( d, 0, -1)
                diodes_shifted = imp.imshift_fft(diodes_ref, shifts[:, 0], shifts[:, 1])
                for ii in range(dc.projections.diode.shape[0]):
                    dc.projections[ii].diode = diodes_shifted[..., ii]
                sino.append(scal*(1-np.moveaxis(diodes_shifted, -1,0)))

                # reconstruct and scale back
                tomo.append( scal * run_sirt( dc, use_absorbances=True, 
                    use_gpu=use_gpu, maxiter=inp.sirt_max_iter, tdqm=False,
                    enforce_non_negativity=True,
                    )['result']['x'][:,:,:,0] )
                
                sys.stdout.write(f'\r\t\t q-value {l*batch_size+k+1:d} / {nq:d}')
                sys.stdout.flush()
            
            # print('\n\tSaving data')
            with h5py.File( out_path, 'r+' ) as hf:
                hf['sinogram'][l*batch_size:(l+1)*batch_size] = np.array(sino)
                hf['tomogram'][l*batch_size:(l+1)*batch_size] = np.array(tomo)
        return out_path

def backproject_qranges( path, path_input, path_data=None, only_mumottize=False ):
    inp = import_module_from_path('input_1Drec', path_input )
    
    with h5py.File( os.path.join(path,'analysis/alignment_result.h5'), 'r' ) as hf:
        kap = hf['kappa'][()]
        ome = hf['omega'][()]
        shifts = hf['shifts'][()]

    kap=np.delete( kap, inp.exclude_proj )
    ome=np.delete( ome, inp.exclude_proj )
    shifts=np.delete( shifts, inp.exclude_proj, axis=0 )
 
    if not path_data:
        path_data = mumottize(path, sub_data=inp.subdir_data, pattern=inp.pattern, 
            q_index_range=[np.min(inp.roi_list), np.max(inp.roi_list)],
            geo_path=os.path.join(path,'analysis/geometry.py'), rec_1d=True,
            )
    
    if not only_mumottize:
        # load mumott file from alignment
        dc = DataContainer(os.path.join(path,'analysis/data_mumott.h5'))

        # set up output file
        title = os.path.basename(os.path.dirname(path))
        out_path = os.path.join(path,'analysis/rec_1d/',
                        f'reconstruction_1Dmm_{title}_{timestring()}.h5')
        with h5py.File( out_path, 'w' ) as hf:
            hf.create_dataset( 'sample_name', data=title )

        # loop over the peaks and reconstruct each q-value
        fov = (dc.projections[0].data.shape[0], dc.projections[0].data.shape[1])
        for q_range, peak_name in zip(inp.roi_list, inp.peak_names):
            print(f'\tReconstructing peak {peak_name}')

            tl = time()
            with h5py.File(path_data, 'r') as hf:
                projections = hf['projections'][
                    q_range[0]-np.min(inp.roi_list):q_range[1]-np.min(inp.roi_list)]
                q_peak = hf['radial_units'][
                    q_range[0]-np.min(inp.roi_list):q_range[1]-np.min(inp.roi_list)]
            print(f'\t\tLoaded data {time()-tl} s')

            nq = projections.shape[0]
            tomograms = np.empty( (fov[0], *fov, nq), data_type)
            sinograms = np.empty( (projections.shape[1], *fov, nq), data_type)
            for k in range( nq ):
                # transform to "diode scale"
                scal = projections[k].max()
                d = 1 - projections[k]/scal

                # insert the data shifted according to alignment
                diodes_ref = np.moveaxis( d, 0, -1)
                diodes_shifted = imp.imshift_fft(diodes_ref, shifts[:, 0], shifts[:, 1])
                for ii in range(dc.projections.diode.shape[0]):
                    dc.projections[ii].diode = diodes_shifted[..., ii]
                sinograms[:,:,:,k] = scal*(1-np.moveaxis(diodes_shifted, -1,0))

                tomograms[:,:,:,k] = scal * run_sirt( dc, use_absorbances=True, 
                    use_gpu=True, maxiter=20, tdqm=False,
                    enforce_non_negativity=True,
                    )['result']['x'][:,:,:,0]
                
                sys.stdout.write(f'\r\t\t q-value {k+1:d} / {nq:d}')
                sys.stdout.flush()

            print('\n\tSaving data')
            with h5py.File( out_path, 'r+' ) as hf:
                gpk = hf.create_group( peak_name )
                gpk.create_dataset( 'radial_units', data=q_peak )
                gpk.create_dataset( 'sinogram', data = sinograms )
                gpk.create_dataset( 'tomogram', data = tomograms )

def backproject( dc, method='sirt', use_abs=False,
                rec_iteration=20, use_gpu=True, tdqm=True ):
    if method=='sirt':
        tomogram = run_sirt( dc, use_absorbances=use_abs, 
            use_gpu=use_gpu, maxiter=rec_iteration, tdqm=tdqm,
        )#['result']['x']    
    else:
        tomogram = run_mitra( dc, use_absorbances=use_abs, 
            use_gpu=use_gpu, maxiter=rec_iteration, tdqm=tdqm,
        )['result']['x']

    # if output:
    #     path_out = os.path.join( os.path.dirname(path_in), 'alignment_result.h5' )
    #     with h5py.File( path_out, 'w') as hf:
    #         hf.create_dataset('shifts', data=np.zeros((dc.diode.shape[0],2),data_type))
    #         hf.create_dataset('sinogram', data=dc.data[:,:,:,0].transpose(1,2,0))
    #         hf.create_dataset('tomogram', data=tomogram )
    #         hf.create_dataset('omega', data=dc.geometry.inner_angles )
    #         hf.create_dataset('kappa', data=dc.geometry.outer_angles )
    #     print(f'Saved alignment result to {path_out}')
    return tomogram

@njit(parallel=True)
def _remove_crazypixels( data ):
    dat_new = data.copy()
    for k in range(data.shape[3]):
        for g in prange(data.shape[0]):
            for l in range(-1, data.shape[1]-1):
                for m in range(-1, data.shape[2]-1):
                    sample = np.array([
                        data[g,l-1,m-1,k],data[g,l-1,m,k],data[g,l-1,m+1,k],
                        data[g,l,m-1,k],                  data[g,l,m+1,k],
                        data[g,l+1,m-1,k],data[g,l+1,m,k],data[g,l+1,m+1,k],
                        ])
                    base = np.median(sample)
                    if data[g,l,m,k] > 10*np.abs(base):
                        dat_new[g,l,m,k] = base
    return dat_new

@njit(parallel=True)
def _average_neighbors(arr, hor=True, ver=True):
    """Numba-compiled function for regrouping data

    Parameters
    ----------
    arr : 2D-array
        data to be regrouped
    hor : bool, optional
        toggle horizontal regrouping, by default True
    ver : bool, optional
        toggle vertical regrouping, by default True

    Returns
    -------
    2D-array
        regrouped data
    """
    rows, cols = arr.shape
    f_row = 2 if ver else 1
    f_col = 2 if hor else 1
    new_rows = (rows + 1) // f_row if ver else rows
    new_cols = (cols + 1) // f_col if hor else cols
    
    # Initialize a new array to hold the averaged values
    result = np.zeros((new_rows, new_cols))

    for i in prange(new_rows):
        for j in range(new_cols):
            # Determine the slice for the 2x2 block
            row_start = f_row * i
            row_end = min(f_row * i + f_row, rows)
            col_start = f_col * j
            col_end = min(f_col * j + f_col, cols)
            
            # Extract the block and compute the average
            block = arr[row_start:row_end, col_start:col_end]
            result[i, j] = np.mean(block)
    
    return result


import numpy as np
from numba import njit, prange

from ..model import rotation as rot
from .. import numba_plugins as nb
from ..analysis import orix_plugins as orx
from ..model import symmetries as sym
from ...config import data_type

def get_odf( coefficients, G_odf, symmetry, resolution ):
    Q_odf = rot.QfromOTP(G_odf)
    Q_grid, Q_group = setup_gridbased(symmetry, resolution)
    return odf( coefficients, Q_odf, Q_grid, Q_group, resolution )


def get_odf_batch( coefficients_batch, G_odf, symmetry, resolution ):
    Q_odf = rot.QfromOTP(G_odf)
    Q_grid, Q_group = setup_gridbased(symmetry, resolution)
    odf_batch = np.empty( (coefficients_batch.shape[0], Q_odf.shape[0]) )
    for c, coefficients in enumerate(coefficients_batch):
        odf_batch[c] = odf( coefficients, Q_odf, Q_grid, Q_group, resolution )

    return odf_batch

@njit
def odf( coefficients, Q_odf, Q_grid, Q_group, std, cutoff=1e-2 ):
    odf = np.zeros_like(Q_odf[:,0])
    for c in range(coefficients.size):
        if coefficients[c] > coefficients.max()*cutoff:
            odf += gaussian_SO3( Q_odf, Q_grid[c], std, Q_group )
            # odf += fisher_SO3( Q_odf, Q_grid[c], kappa, Q_group)
    return odf

def setup_gridbased( symmetry, resolution ):
    pg = getattr( orx.osym, sym.get_SFnotation( symmetry ) )
    Q_grid = rot.get_sample_fundamental(
                    resolution,
                    point_group= pg,
                    method='cubochoric'
            ).data.astype(data_type)
    
    # Q_grid = rot.QfromOTP(np.array([[0,0,0], [np.pi/2,np.pi/2,np.pi/2]]))
    # resolution = 10

    generators = sym.generators(symmetry)
    q_gen = rot.QfromOTP(generators)
    Q_group = rot.generate_group(q_gen)

    return Q_grid, Q_group

@njit(parallel=True)
def get_odf_max_orientations( Coefficients, Q_odf, Q_grid, Q_group, kappa ):
    Q_max = np.empty( (Coefficients.shape[0], 4), data_type )
    for v in prange(Coefficients.shape[0]):
        odf = odf( Coefficients[v], Q_odf, Q_grid, Q_group, kappa )
        Q_max[v] = Q_odf[np.argmax(odf)]
    return Q_max

@njit
def fisher_SO3(Q, q_mean, kappa, Q_group): 
    """
    von Mises-Fisher distribution on fundamental zone (fz)

    Parameters
    --------
    Q : 2d ndarray, float
        array of unit quaternions representing orientations
    q_mean: 1d ndarray, float 
        mean orientation as quaternion
    kappa: float
        concentration parameter for von Mises-Fisher distribution (~1/sigma^2) ! this value seems not correct
    Q_group: 2d ndarray, float
        array of unit quaternions with all symmetry operations of the point group
    
    Returns:
    ------------
    odf: 1d ndarray, float
        non-normalized probabilty density for each quaternion
    """
    # We put multiple (mises-fisher) bell functions on the unit quaternion sphere
    # One on every symmetry equivalent q_mu
    # Evaluate the sum of these on orientations Q
    q_mean_equivalents = rot.symmetry_equivalent_quaternions(q_mean,Q_group)
    # q_mu_equivalents = np.atleast_2d(q_mu) # this is to show what happens if you use only q_mu

    # calculate odf as a sum of distributions from equivalents
    odf = np.zeros((q_mean_equivalents.shape[0],Q.shape[0]), dtype=data_type)
    for i in range(q_mean_equivalents.shape[0]):
        mux = np.abs( q_mean_equivalents[i] @ Q.T )
        odf[i]= np.exp(kappa * (mux - 1))
    odf = np.sum(odf, axis=0)# / np.exp(kappa)
    return odf

@njit
def gaussian_SO3(Q, q_mean, std, Q_group): 
    """
    Normal-like distribution

    Parameters
    --------
    Q : 2d ndarray, float
        array of unit quaternions representing orientations
    q_mean: 1d ndarray, float 
        mean orientation as quaternion
    std: float
        standard deviation in radians
    Q_group: 2d ndarray, float
        array of unit quaternions with all symmetry operations of the point group
    
    Returns:
    ------------
    odf: 1d ndarray, float
        non-normalized probabilty density for each quaternion
    """
    
    dQ = rot.misorientation_angle_stack(Q, nb.nb_full(Q.shape,q_mean), Q_group)
    odf = np.exp( - dQ**2/(2*std**2) )

    return odf

@njit(parallel=True)
def get_rotated_diffractlets(Qc, Qs, Q_grid, Q_group, kappa, I_single_crystal, detShape ):
    difflets_rot = np.empty( (Qs.shape[0], Q_grid.shape[0], *detShape), data_type )
    # s,gr=0,0
    for s in prange(Qs.shape[0]):
        for gr in range(Q_grid.shape[0]):
            q_mean = rot.quaternion_multiply( Qs[s], Q_grid[gr] )
            odf = fisher_SO3( Qc, q_mean, kappa, Q_group )
            # sparse calculate the projections (only points in odf that are high)
            diff_pattern = np.zeros(detShape, data_type)
            idcs_odf = np.nonzero( odf > 0.01 * odf.max() )[0] # cut small values of the odf
            for h in idcs_odf:
                diff_pattern += I_single_crystal[h] * odf[h]
            difflets_rot[s,gr] = diff_pattern
    return difflets_rot





# @njit
# def gaussian_3d( Q, q_mu, std, gen, dV=1 ):
#     """
#     Gauss bell on FZ
#     Parameters
#     --------
#     Q : 2d ndarray, float
#         array of unit quaternions representing orientations
#     mu: 1d ndarray, float 
#         mean orientation as quaternion
#     std: float
#         standard deviation - sigma
#     gen: 2d ndarray, float
#         OTP of the two generators for the point group symmetries
#         dim0: generators, dim1: OTP
    
#     Returns:
#     ------------
#     odf: 1d ndarray, float
#         non-normalized probabilty (mass) for each of the orientations g
#     """
#     # for omega_mu = 0, then dg is omega
#     dg = rot.ang_distance(Q, nb.nb_full(Q.shape,q_mu), gen)
#     odf = np.exp( - dg**2/(2*std**2) )
#     return odf #/( odf @ dV ) Does not return a valid pmf at the moment

# from .misc import integrate_c
# from numba import prange
# # @njit(parallel=True)
# def projection( g, Qc, Isc, gen, Q_mu, c_sample, kappa, Qs, Beams, iBeams, detShape, dV ):
#     diff_patterns_g = np.empty((Beams.shape[1],detShape[0],detShape[1]), data_type)
#     for t in prange(Beams.shape[1]):
#         # project the coefficients
#         iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
#         c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], c_sample )
#         # get the resulting odf from rotated mu
#         odf_proj = np.zeros( Qc.shape[0], data_type )
#         idcs_basis = np.nonzero(c_proj > 0.01 * c_proj.max())[0]
#         for c in idcs_basis:
#             q_mu = rot.quaternion_multiply(  Q_mu[c], Qs[g] )
#             odf_proj += c_proj[c] * fisher_SO3(Qc, q_mu, kappa, gen, dV )
#         # sparse calculate the projections (only points in odf that are high)
#         diff_pattern = np.zeros(Isc.shape[1], data_type)
#         idcs_odf = np.nonzero(odf_proj> 0.01 * odf_proj.max())[0]
#         for h in idcs_odf:
#             diff_pattern += Isc[h] * odf_proj[h]
#         diff_patterns_g[t] = diff_pattern.reshape((detShape[0],detShape[1]))
#     return diff_patterns_g






import numpy as np
from numba import njit

from ..config import data_type

@njit()
def integrate_c(sparsearray, indices, C ):
    """ Sparse-multiplication for several entries at once 

    Parameters
    ----------
    sparsearray : ndarray, float
        list of the non-zero entries
    indices : ndarray, int
        indices of the non-zero entries
    C : 2D ndarray, float
        conventional array
        dim 0: to be summed over, 1: whatever
    
    returns: 
    ------------
    c_int : ndarray float
        vector product for every entry in dim 1 of C
    """
    c_int = np.zeros( C.shape[1], C.dtype)
    for i in range(indices.shape[0]):
        idx = indices[i]
        c_int += C[idx] * sparsearray[i] # get weighted average hsh-coefficients
    return c_int

@njit()
def sparsemult(sparsearray, indices, vector):
    """ Vector product using one custom sparse-array

    Parameters
    ----------
    sparsearray : ndarray, float
        list of the non-zero entries
    indices : ndarray, int
        indices of the non-zero entries
    vector : ndarray, float
        conventional vector
    
    returns: 
    ------------
    sum : float
    """
    sum = 0.
    for i,S in zip(indices, sparsearray):
        sum += vector[i] * S
    return sum

@njit
def masked_mean_axis0(data, mask):
    """
    Column-wise mean (axis=0) of data, including only elements where mask != 0.
    data, mask must have same shape.
    """
    n_rows, n_cols = data.shape
    out = np.empty(n_cols, dtype=data.dtype)

    for j in range(n_cols):
        s = 0.0
        count = 0
        for i in range(n_rows):
            if mask[i, j] != 0:
                s += data[i, j]
                count += 1
        out[j] = s / count if count > 0 else 0.0
    return out

@njit
def masked_sum_axis1(data, mask):
    """
    Row-wise sum (axis=1) of data, including only elements where mask != 0.
    data, mask must have same shape.
    """
    n_rows, n_cols = data.shape
    out = np.empty(n_rows, dtype=data.dtype)

    for i in range(n_rows):
        s = 0.0
        for j in range(n_cols):
            if mask[i, j] != 0:
                s += data[i, j]
        out[i] = s
    return out

@njit
def masked_average_data(data,mask):
    """averages each image while masking

    Parameters
    ----------
    data : 3d array, float
        0: stack of images, 1: chi, 2: q
    mask : 2darray, bool
        mask for each image

    Returns
    -------
    1darray, float
        average over each image
    """
    out = np.empty_like( data[:,0,0] )
    for i in range(data.shape[0]):
        s = 0.0
        count = 0
        for j in range(data.shape[1]):
            for k in range(data.shape[2]):
                if mask[j, k] != 0:
                    s += data[i, j, k]
                    count += 1
        out[i] = s / count if count > 0 else 0.0
    # for i in range(data.shape[0]):
    #     out[i] = data[i][mask].mean()
    return out 

@njit
def trapz(y, dx):
    """
    Numerical integration using the trapezoidal rule.

    Parameters
    ----------
    y : 1D array
        Function values at points x.
    dx : float
        Increment in x

    Returns
    -------
    float
        Approximate integral of y with respect to x.
    """
    n = len(y)
    total = 0.0
    for i in range(n - 1):
        total += 0.5 * (y[i] + y[i+1]) * dx
    return total

@njit()
def nb_meshgrid(x, y):
    """Numpy-like meshgrid function that is compatible with numba

    Parameters
    ----------
    x,y : ndarray, float

    returns: 
    ------------
    xx,yy : 2D ndarray, float
    """
    xx = np.empty(shape=(y.size, x.size), dtype=x.dtype)
    yy = np.empty(shape=(y.size, x.size), dtype=y.dtype)
    for j in range(y.size):
        for k in range(x.size):
            xx[j,k] = x[k]  # change to x[k] if indexing xy
            yy[j,k] = y[j]  # change to y[j] if indexing xy
    return xx, yy

@njit()
def nb_erf(x):
    ''' gives a numeric approximation of the error-function '''
    # save the sign of x
    sign = np.sign(x)
    x = np.abs(x)

    # constants
    a1 =  0.254829592
    a2 = -0.284496736
    a3 =  1.421413741
    a4 = -1.453152027
    a5 =  1.061405429
    p  =  0.3275911

    # A&S formula 7.1.26
    t = 1.0/(1.0 + p*x)
    y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*np.exp(-x*x)
    return sign*y # erf(-x) = -erf(x)

@njit
def nb_vectornorm(v):
    """Gives the vectornorm along the first axis
    """
    norm = np.sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2])
    return norm

    # if norm == 0.0:
    #     return v
    # return v / norm

@njit
def nb_dot(u, v):
    """Gives the dot product along the first axis
    """
    return u[0]*v[0] + u[1]*v[1] + u[2]*v[2]

@njit
def nb_cross(u, v):
    """Gives the cross product along the first axis
    """
    return np.array([
        u[1]*v[2] - u[2]*v[1],
        u[2]*v[0] - u[0]*v[2],
        u[0]*v[1] - u[1]*v[0],
    ])

@njit
def nb_polyfit(x, y, degree):
    """
    Fit a polynomial of a specified degree to the given data using least squares approximation.
    Numba compatible version of the numpy function
    
    Parameters:
    x (array-like): The x-coordinates of the data points.
    y (array-like): The y-coordinates of the data points.
    degree (int): The degree of the polynomial to fit.

    Returns:
    array: Coefficients of the polynomial, highest power first.
    """
    n = len(x)
    m = degree + 1
    A = np.empty((n, m), data_type)
    for i in range(n):
        for j in range(m):
            A[i, j] = x[i] ** (degree - j)
    return np.linalg.lstsq(A, y )[0]

@njit
def nb_polyval(coeff, x):
    """
    Evaluate a polynomial given the coefficients at the points x.
    Uses Horner's Method.
    Numba compatible version of the numpy function

    Parameters:
    coeff (array-like): The coefficients of the polynomial.
    x (array-like): The x-coordinates of the data points.
    """
    res = np.zeros_like(x)
    for c in coeff:
        res = x * res + c
    return res

@njit
def nb_tile_1d(a, n):
    # numba-optimized function to c
    # Create an output array of the desired shape
    out = np.empty((n, len(a)), data_type)
    
    # Fill the output array with repeated values from a
    for i in range(n):
        out[i] = a
    
    return out

@njit
def nb_mean_ax0(a):
    # numba-optimized function to calculate mean values along the first dimension
    res = np.empty( a.shape[1], data_type)
    for i in range(a.shape[1]):
        res[i] = a[:,i].mean()
    return res

@njit
def nb_isnan(array):
    # Create a boolean mask where True indicates NaN values
    mask = np.zeros(array.shape, dtype=np.bool_)
    for i in range(array.shape[0]):
        for j in range(array.shape[1]):
            for k in range(array.shape[2]):
                # Check if the value is NaN by comparing the value to itself
                if array[i, j, k] != array[i, j, k]:
                    mask[i, j, k] = True
    return mask

@njit
def nb_clip_array(x, min_val, max_val):
    """Numba-compatible equivalent of np.clip."""
    x[x<min_val] = min_val
    x[x>max_val] = max_val
    return x

@njit
def nb_clip(x, min_val, max_val):
    """Numba-compatible equivalent of np.clip."""
    if x < min_val:
        return min_val
    elif x > max_val:
        return max_val
    return x

@njit
def nb_unique_axis0(arr):
    n, m = arr.shape
    output = np.empty((n, m), dtype=arr.dtype)
    count = 0

    for i in range(n):
        duplicate = False
        for j in range(count):
            is_same = True
            for k in range(m):
                if arr[i, k] != output[j, k]:
                    is_same = False
                    break
            if is_same:
                duplicate = True
                break
        if not duplicate:
            for k in range(m):
                output[count, k] = arr[i, k]
            count += 1

    return output[:count]

@njit
def nb_full(shape, fill_array):
    out = np.empty((shape[0], fill_array.shape[0]), dtype=fill_array.dtype)
    for i in range(shape[0]):
        for j in range(fill_array.shape[0]):
            out[i, j] = fill_array[j]
    return out


# textom/src/optimization/__init__.py
import pkgutil
import importlib

# Dynamically import all submodules and expose their contents
__all__ = []
for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    module = importlib.import_module(f"{__name__}.{module_name}")
    for name in dir(module):
        if not name.startswith("_"):  # Skip private attributes
            globals()[name] = getattr(module, name)
            __all__.append(name)




# choose a mode among the following
mode = 'linear' # options: 'linear' # 'chebyshev' # 'linear_azimuthal' # 'chebyshev_azimuthal' #'none' # 'custom'
# not used parameters will be ignored
order_chebyshev = 10 
# auto masking parameters ('chebyshev_auto')
pre_order=6 # chebyshev order for auto-masking (to smooth the curve)
k_sigma=2.0 # this measures how much the peak stands out from the data
peak_expand=2 # this should be about the width of the peaks
# choose q-range where the baseline should be drawn (can be important for 'chebyshev_auto' & 'chebyshev')
q_min=0 
q_max=100
# custom mode
path_custom_bl = 'analysis/custom_baselines.h5' # path relative to sample_dir
h5path_custom_bl = 'baselines' # h5 internal path


'''
make another option for if you know we have fibre textures (maybe also make platelet)
the problem reduces to 2D -> SH

sth like this:
def get_odf( Q, v, odf_2D ):

    odf = np.zeros_like(Q[:,0])
    for q in range(odf):
        z_vector = rot.quaternion_rotate_vector(Q(q), np.array([0,0,1]))
        dot = np.dot(z_vector, v) # maybe sparsify, maybe gaussian like Mads
        odf[q] = np.sum( dot * odf_2D )

'''


''' fitting several polymorphs at a time
could wrap several fit objects

how do i organise files?
    difflets
    data_textom - needs to be only one actually. if peaks overlap they go together
    peak_regions (baseline regions will be common)
    crystal.py


optimizer
    this one should be given just a loss function and a gradient, which i can put together in multifit, that works with a
    combined list of coefficients

load_opt might be hard

visualization (basically all functions that depend on fit)
    all functions should be preserved if there is several, i could just make the fit object a list
    and then functions get an argument that choses the polymorph, with default 0

results
    this is maybe hard, need to label everything with the polymorph number?

'''


import numpy as np
import os
import sys
import shutil
import importlib
import inspect
from datetime import datetime
import sys
import time
import re
import glob
import platform

from .handle import get_file_path
from ..version import __version__
from ..config import fun_mode

crystal_ascii = r"""
    ____     ____  
   /\   \   /\   \ 
  /__\___\ /__\___\
  \  / TexTOM /   /
   \/___/   \/___/ 
"""

def fancy_title():
    # # Clear the terminal
    os.system('cls' if os.name == 'nt' else 'clear')

    if fun_mode:
        intro = add_phrase_to_crystal()
    else:
        intro = crystal_ascii

    # Display the title gradually as an animation
    for char in intro:
        sys.stdout.write(char)
        sys.stdout.flush()
        time.sleep(0.0011)  # Adjust the speed of the animation
    time.sleep(0.03)
    # Add a final message
    print("\n\nWelcome to TexTOM!")
    print("====================================")
    print("A program for texture simulations.")
    print("\u00A9 Moritz Frewein, Moritz Stammer, Marc Allain, Tilman Gr\u00FCnewald.")
    print(f"Version {__version__}")
    print("====================================")
    time.sleep(0.05)

def add_phrase_to_crystal():
    line = read_phrase()
    words = line.split(' ')
    line1 = ' '.join(words[:len(words)//2])
    line2 = ' '.join(words[len(words)//2:])
    cr = crystal_ascii.split('\n')
    cr[2] = cr[2] + '    | ' + '-'*max(len(line1),len(line2))
    cr[3] = cr[3] + '    | ' + line1
    cr[4] = cr[4] + '    | ' + line2
    cr[5] = cr[5] + '    | ' + '-'*max(len(line1),len(line2))
    return '\n'.join(cr)

def read_phrase():
    phrases_path = get_file_path('textom',
            os.path.join('ressources','wise_phrases.txt') )
    with open(phrases_path) as fid:
        lines = fid.readlines()
    return lines[np.random.randint(len(lines))][:-1]

def check_version():

    return 0

def cp_addno(source_file, destination_folder):
    """ copies a file to destination. if a file with the same
    name exits, adds a number to the name"""
    # Check if the source file exists
    if not os.path.isfile(source_file):
        print("Source file does not exist.")
        return

    # Get the filename and extension
    filename, extension = os.path.splitext(os.path.basename(source_file))

    # Prepare the destination path
    destination_file = os.path.join(destination_folder, os.path.basename(source_file))

    # If the file exists at the destination, add a number to the filename
    if os.path.exists(destination_file):
        index = 1
        while True:
            new_filename = f"{filename}_{index}{extension}"
            new_destination_file = os.path.join(destination_folder, new_filename)
            if not os.path.exists(new_destination_file):
                destination_file = new_destination_file
                break
            index += 1

    # Copy the file to the destination
    shutil.copyfile(source_file, destination_file)
    print(f"File copied to: {destination_file}")

def cp_add_dt(source_file, destination_folder, now=True):
    """ copies a file to destination and attaches a timestamp in the title"""
    # Check if the source file exists
    if not os.path.isfile(source_file):
        print("Source file does not exist.")
        return

    os.makedirs( destination_folder, exist_ok=True )

    # Get the filename and extension
    filename, extension = os.path.splitext(os.path.basename(source_file))
    # choose between current time and file modification time
    dt = timestring() if now else timestring(source_file) 
    # Prepare the destination path
    destination_file = os.path.join(destination_folder, 
                            f'{filename}_{dt}{extension}')

    # Copy the file to the destination
    shutil.copyfile(source_file, destination_file)
    print(f"File copied to: {destination_file}")

def import_module_from_path(module_name, file_path):
    # Create a module spec from the file path
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    
    # Create a new module based on the spec
    module = importlib.util.module_from_spec(spec)
    
    # Add the module to the system modules
    sys.modules[module_name] = module
    
    # Execute the module in its own namespace
    spec.loader.exec_module(module)
    
    return module

def print_library_functions_and_docstrings(module, filter='', sort_alphabetical=False):
    """
    Prints all functions of a library and the first line of their docstrings.

    Parameters:
    module (module): The library module to inspect.
    filter (str): Optional filter to only show functions containing this substring.
    sort_alphabetical (bool): If True, functions are printed in alphabetical order.
                              If False, functions are printed in the order they appear in the library.
    """  
    print('Available methods:')
    # Retrieve all functions in the module, including imported ones
    functions = [
        (name, obj) for name, obj in ((name, getattr(module, name)) for name in dir(module))
        if inspect.isfunction(obj) and name[0] != '_' and filter in name
        and inspect.getmodule(obj) == module
    ]
    # functions = [
    #     (name, obj) for name, obj in inspect.getmembers(module)
    #     if inspect.isfunction(obj) and name[0] != '_' and filter in name
    # ]
    # # Determine the order of functions based on source code if sort_alphabetical is False
    # if not sort_alphabetical:
    #     try:
    #         source = inspect.getsource(module)
    #         function_order = re.findall(r'def\s+([a-zA-Z_][a-zA-Z0-9_]*)', source)
    #         function_order = [name for name in function_order if filter in name]
    #         functions_dict = dict(functions)
    #         functions = [(name, functions_dict[name]) for name in function_order if name in functions_dict]
    #     except (OSError, TypeError):
    #         print("Could not retrieve source code. Showing in alphabetical order.")
    
    # Sort functions alphabetically if sort_alphabetical is True
    if sort_alphabetical:
        functions = sorted(functions, key=lambda x: x[0])

    # Print function names and their first docstring line
    for name, obj in functions:
        docstring = inspect.getdoc(obj)
        first_line = docstring.split('\n')[0] if docstring else 'No docstring available'
        print(f"\t{name} : {first_line}")

def list_library_functions(module):
    """
    Lists all functions of a library.

    Parameters:
    module (module): The library module to inspect.

    Returns:
    List[str]: A list of function names in the module.
    
    Example:
    functions = list_library_functions(numpy)
    print(functions)
    """
   
    # Initialize an empty list to store function names
    function_names = []
    
    # Iterate through the members of the module
    for name, obj in inspect.getmembers(module):
        if inspect.isfunction(obj):
            # Append the function name to the list
            function_names.append(name)
    
    return function_names

def print_function_docstring(module, function_name):
    """
    Prints the docstring of a specified function in a module.

    Parameters:
    module (module): The library module to inspect.
    function_name (str): The name of the function whose docstring is to be printed.
    
    Example:
    print_function_docstring(math, 'sqrt')
    """
   
    # Get the function object from the module
    function = getattr(module, function_name, None)
    
    if function and inspect.isfunction(function):
        # Get the docstring of the function
        docstring = inspect.getdoc(function)
        if docstring:
            print(f"Function {function_name}:\n{docstring}")
        else:
            print(f"No docstring available for {function_name}.")
    else:
        print(f"{function_name} is not a valid function in the given module.")

def timestring( file_path=None ):
    """Makes a timestamp string for convenient for saving data

    If given a filepath, takes the modification datetime, else the current moment
    
    Parameters
    ----------
    file_path : str
    
    Returns
    -------
    str
        timestamp e.g. '2024_11_22_15h37'
    """
    if file_path:
        dt = datetime.fromtimestamp( os.path.getmtime(file_path) )
    else:
        dt = datetime.now()
    return str(dt).replace(' ','_').replace('-','_').replace(':','h')[:-10]

def lastfile( path, filter='*' ):
    """Searches for the youngest file in the path an returns the name

    Parameters
    ----------
    path : str
        directory containing the files
    filter : str
        searches for all file corresponding to pattern, by default '*'

    Returns
    -------
    str
        name of the youngest file
    """
    search_pattern = os.path.join( path, filter )
    h5_files = glob.glob(search_pattern)
    if any(h5_files):
        h5path = max(h5_files, key=os.path.getmtime)
        return h5path
    else:
        print('No files found')

def crop_to_non_nan_region(array):
    """
    Crop a 2D array to the smallest rectangular region that contains no NaN values.
    
    Parameters:
    - array: 2D numpy array with potential NaN values.
    
    Returns:
    - Cropped 2D numpy array containing only the region with no NaNs.
    """
    # Find the indices of non-NaN values
    non_nan_rows = np.any(~np.isnan(array), axis=1)
    non_nan_cols = np.any(~np.isnan(array), axis=0)
    
    # Get the bounding box for rows and columns
    row_start, row_end = np.where(non_nan_rows)[0][[0, -1]]
    col_start, col_end = np.where(non_nan_cols)[0][[0, -1]]
    
    # Crop the array to the non-NaN region
    cropped_array = array[row_start:row_end+1, col_start:col_end+1]
    
    return cropped_array, [row_start, row_end, col_start, col_end]

def smallest_subvolume(field):
    """
    Finds the smallest subvolume containing all nonzero vectors in a 3D field or vector field.

    Parameters:
    field (np.ndarray): A 3D or 4D array of shape (nx, ny, nz, 3), where the last dimension
                        represents the vector components (x, y, z).

    Returns:
    tuple: (sliced_field, slices), where:
        - sliced_field is the smallest subvolume containing all nonzero vectors.
        - slices is a tuple of slice objects for the original array.
    """
    # Check for nonzero vectors
    if field.ndim == 4:
        nonzero_mask = np.any(field != 0, axis=-1)  # True for nonzero vectors
    elif field.ndim == 3:
        nonzero_mask = ~np.isnan(field) & (field != 0)

    # Find the bounds along each axis
    nonzero_indices = np.where(nonzero_mask)
    x_min, x_max = np.min(nonzero_indices[0]), np.max(nonzero_indices[0])
    y_min, y_max = np.min(nonzero_indices[1]), np.max(nonzero_indices[1])
    z_min, z_max = np.min(nonzero_indices[2]), np.max(nonzero_indices[2])

    # Define slices for the subvolume
    slices = (slice(x_min, x_max + 1), slice(y_min, y_max + 1), slice(z_min, z_max + 1))

    # Extract the subvolume
    subvolume = field[slices]

    return subvolume, [x_min,x_max,y_min, y_max,z_min, z_max]

def find_indices_in_range(array, value_range):
    """
    Finds the index range of entries in a 1D array within a specified range.

    Parameters:
    array (np.ndarray): A 1D array with steadily increasing values.
    value_range (tuple): A tuple (lower, upper) specifying the range of values.

    Returns:
    tuple: (lower_index, upper_index), where:
        - lower_index is the index of the first value >= lower.
        - upper_index is the index of the first value >= upper.
    """
    lower, upper = value_range

    # Use searchsorted to find insertion indices
    lower_index = np.searchsorted(array, lower, side='left')
    upper_index = np.searchsorted(array, upper, side='right')# - 1

    # Ensure indices are within bounds
    if lower_index >= len(array) or array[lower_index] < lower:
        lower_index = -1  # No valid lower index
    if upper_index < 0 or array[upper_index-1] > upper:
        upper_index = -1  # No valid upper index

    return lower_index, upper_index

def find_unique_positions( positions ):
    """This assumes that positions are on a 1d grid but there is noise.

    Finds out the grid points and puts them in ascending order.
    Categorizes the positions into the grid points and returns the indices in the grid.

    Parameters
    ----------
    positions : ndarray, float

    Returns
    ----------
    indices: 1darray, int
        corresponding indices for each position to the grid
    """
    positions = np.array(positions).flatten()
    diffs = np.diff(np.sort(positions))
    if diffs.max() > 5*diffs.min(): # if there is several points at the same grid point
        threshold = np.median(diffs) + np.std(diffs)
        step = np.median(diffs[diffs > threshold])  # ignore steps that come from noise
        a0 = np.min(positions)
        # Quantize each value to its nearest grid index
        indices = np.round((np.array(positions) - a0) / step).astype(int)
        # unique_idx = np.unique(idx)
    else:
        # only one z-position
        indices = np.argsort(positions) 
    return indices


def list_files_by_modification_time(directory, reverse=False, extension=None):
    """
    List all files in a directory ordered by modification time, optionally filtered by file extension.

    Parameters:
    directory (str): Path to the directory.
    reverse (bool): If True, list files from newest to oldest. Default is False (oldest to newest).
    extension (str): Optional file extension to filter by (e.g., '.txt'). Default is None (no filtering).

    Returns:
    list: A list of filenames in the directory, ordered by modification time and filtered by extension if provided.
    """
    # Get all files in the directory
    files = [f for f in os.listdir(directory) if os.path.isfile(os.path.join(directory, f))]
    
    # Filter by extension if provided
    if extension:
        files = [f for f in files if f.lower().endswith(extension.lower())]
    
    # Sort files by modification time
    files_sorted = sorted(files, key=lambda x: os.path.getmtime(os.path.join(directory, x)), reverse=reverse)
    
    return files_sorted

def insert_sparse_tomogram( dimensions, mask, values ):
    """ Makes a 3D tomogram out of sparse data
    """
    try:
        tomogram = np.empty( (dimensions.prod(),values.shape[1]), np.float64 )
        tomogram[:] = np.nan
        tomogram[mask] = values
        tomogram = tomogram.reshape((*dimensions, values.shape[1]))
    except:
        tomogram = np.empty( dimensions.prod(), np.float64 )
        tomogram[:] = np.nan
        tomogram[mask] = values
        tomogram = tomogram.reshape(dimensions)
    return tomogram

def is_running_in_jupyter():
    """
    Checks if the code is running inside a Jupyter notebook.
    Returns:
        bool: True if running in Jupyter, False otherwise.
    """
    try:
        from IPython import get_ipython
        if 'ipykernel' in sys.modules and isinstance(get_ipython(), type(None)) == False:
            return True
    except ImportError:
        pass
    return False

def get_affinity_info():
    """this is to check if the process is runnin under taskset with restricted CPUs
    """
    if platform.system() == "Linux":
        allowed = sorted(os.sched_getaffinity(0))
        return {
            "restricted": len(allowed) < os.cpu_count(),
            "allowed_cores": allowed,
            "n_cores": len(allowed)
        }
    else:
        # macOS / Windows → no affinity API (no taskset)
        return {
            "restricted": False,
            "allowed_cores": None,
            "n_cores": os.cpu_count()
        }


import numpy as np
## Define diffraction-related parameters:
# x-ray energy in keV
E_keV = 50
# q range for fitting: (lower,upper) boundary in nm^-1
q_range = (10,60)
# path to crystal cif file
cifPath = '/Users/moritz/Documents/code/modelling_cajetan/scripts_cajetan/CIF-Files/ti64_alpha_modified.cif'
# approximate crystal size along (a,b,c)-axes in nm
crystalsize = (15,15,15)
# angular sampling
sampling = 'cubochoric' # or 'simple' (legacy)
# angular sampling resolution
dchi = 2*np.pi / 120





from scipy.ndimage import map_coordinates
from numba import njit, prange
import numpy as np

from textom.src.model import rotation as rot
from textom.src.model.model_textom import model_textom

from textom.input import geometry as geo
def all_paths( mod:model_textom ):
    
    for g in range(mod.Gs.shape[0]):
        beam_direction = geo.beam_direction @ rot.MatrixfromOTP(mod.Gs[g,0],mod.Gs[g,1],mod.Gs[g,2])
        detector_direction_origin = geo.detector_direction_origin @ rot.MatrixfromOTP(mod.Gs[g,0],mod.Gs[g,1],mod.Gs[g,2])
        detector_direction_positive_90 = geo.detector_direction_positive_90 @ rot.MatrixfromOTP(mod.Gs[g,0],mod.Gs[g,1],mod.Gs[g,2])

        mod.x_p
        

def integrate_mu(mu, point_start, point_end, step):
    direction = point_end - point_start
    length = np.linalg.norm(direction)
    direction = direction / length
    t_vals = np.arange(0, length/step, step)
    points = point_start[:, None] + direction[:, None] * t_vals
    mu_vals = map_coordinates(mu, points, order=1, mode='nearest')
    return np.sum(mu_vals) * step

@njit(parallel=True)
def integrate_detector_pixel(tomogram, beam_dir, two_theta, chi, stepsize):
    nx, ny, nz = tomogram.shape
    kf = np.array([ # this needs to be adapted to beam_dir (actually instead of beam_dir, i need the sample orientation as input and then i back-rotate both k_0 and k_scattered)
        np.sin(two_theta) * np.cos(chi),
        np.sin(two_theta) * np.sin(chi),
        np.cos(two_theta)
    ])
    transmission = np.zeros(tomogram.size)
    idx = 0
    for ix in prange(nx): # here i need to apply the scanmask
        for iy in range(ny):
            for iz in range(nz):
                point = np.array([ix, iy, iz])
                A_in = integrate_along_vector(tomogram, point - 100*beam_dir, point, stepsize)
                A_out = integrate_along_vector(tomogram, point, point + 100*kf, stepsize)
                transmission[idx] = np.exp(-(A_in + A_out))
                idx += 1
    return np.sum(transmission)
'''
basically need to do this:
for s along primary beam:
    point = beam_origin + s*beam_dir
    A_in = integral of μ from beam start to point
    A_out = integral over cone paths from point to detector pixel
    contribution = exp(-A_in - A_out)
    accumulate
'''

@njit
def integrate_along_vector(array_3d, point_start, point_end, stepsize):
    """
    Numerically integrates values from a 3D array along a straight path
    between two points using trilinear interpolation.

    Parameters
    ----------
    array_3d : 3D numpy array
        Field of absorption coefficients or scalar values on a regular grid.
    point_start : array-like of float (3,)
        Starting point of the integration path, in voxel coordinates.
    point_end : array-like of float (3,)
        Ending point of the integration path, in voxel coordinates.
    stepsize : float
        Distance between consecutive interpolation samples (in voxel units).

    Returns
    -------
    float
        Approximate line integral of array_3d along the path.
    """
    # Direction vector and total path length
    direction = point_end - point_start
    length = np.sqrt(direction[0]**2 + direction[1]**2 + direction[2]**2)
    inv_len = 1.0 / length
    dx, dy, dz = direction * inv_len

    n_steps = int(length / stepsize)

    nx, ny, nz = array_3d.shape
    ox, oy, oz = point_start
    integral = 0.0

    for s in range(n_steps):
        # Current position along the path
        x = ox + dx * s * stepsize
        y = oy + dy * s * stepsize
        z = oz + dz * s * stepsize

        # Integer voxel indices
        ix, iy, iz = int(x), int(y), int(z)

        # Skip points outside valid interpolation range
        if 0 <= ix < nx-1 and 0 <= iy < ny-1 and 0 <= iz < nz-1:
            # Fractional offsets within voxel
            fx, fy, fz = x - ix, y - iy, z - iz

            # Trilinear interpolation
            c000 = array_3d[ix, iy, iz]
            c100 = array_3d[ix+1, iy, iz]
            c010 = array_3d[ix, iy+1, iz]
            c001 = array_3d[ix, iy, iz+1]
            c110 = array_3d[ix+1, iy+1, iz]
            c101 = array_3d[ix+1, iy, iz+1]
            c011 = array_3d[ix, iy+1, iz+1]
            c111 = array_3d[ix+1, iy+1, iz+1]

            c00 = c000*(1-fx) + c100*fx
            c01 = c001*(1-fx) + c101*fx
            c10 = c010*(1-fx) + c110*fx
            c11 = c011*(1-fx) + c111*fx
            c0 = c00*(1-fy) + c10*fy
            c1 = c01*(1-fy) + c11*fy
            interpolated_value = c0*(1-fz) + c1*fz

            # Accumulate integral
            integral += interpolated_value

    # Scale by step size to approximate continuous integral
    return integral * stepsize


import numpy as np # don't delete this line
##################################################

# Define how many cores you want to use 
# If this number is higher than the available cores on your system, 
# it will fall back to the maximum number
n_threads = 128 

# Choose if you want to use a GPU for alignment
use_gpu = True
# needs cudatoolkit: pip install cudatoolkit

# Choose your precision
# recommended np.float64 for double or np.float32 for single precision
# this mainly concerns data handling, critical parts of the code always use double precision
data_type = np.float32

# for visualization:
odf_resolution = 3 # degree
hex_notation = True # set to True to use 4-axis Miller-Bravais indices for hexagonal lattices

# turn on wise phrases at the start of TexTOM
fun_mode = False


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib.patches import Rectangle
from matplotlib.widgets import Button, TextBox
from skimage.measure import marching_cubes
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from scipy import ndimage
plt.ion()

""" Library for creating masks """
####################################################
class RegionSelector:
    def __init__(self, ax, ymin, ymax, max_regions=None, peaks=None):
        """
        Initialize the RegionSelector class.

        Parameters:
        ax (matplotlib.axes.Axes): The Axes object where the data is plotted.
        ymin (float): minimum value of the data.
        ymax (float): maximum value of the data.
        max_regions (int, optional): The maximum number of regions to select. If None, an arbitrary number of regions can be selected.
        """
        self.ax = ax
        self.ymin = ymin
        self.ymax = ymax
        self.max_regions = max_regions
        self.press = None
        self.rects = []
        self.current_rect = None
        self.peaks = peaks # this is to put another color if no peak is contained

        self.cidpress = ax.figure.canvas.mpl_connect('button_press_event', self.on_press)
        self.cidrelease = ax.figure.canvas.mpl_connect('button_release_event', self.on_release)
        # self.cidmotion = ax.figure.canvas.mpl_connect('motion_notify_event', self.on_motion)
        self.cidclick = ax.figure.canvas.mpl_connect('button_press_event', self.on_rightclick)

    def on_press(self, event):
        """
        Handle the mouse press event.

        Parameters:
        event (matplotlib.backend_bases.MouseEvent): The mouse event.
        """
        if event.inaxes != self.ax: return
        if self.ax.figure.canvas.toolbar.mode == 'zoom rect': return # does nothing when zooming
        if event.button != 1: return  # Only respond to left mouse button
        if self.max_regions is not None and len(self.rects) >= self.max_regions: return

        self.press_x = event.xdata
        self.ax.figure.canvas.draw()

    def on_release(self, event):
        """
        Handle the mouse release event.

        Parameters:
        event (matplotlib.backend_bases.MouseEvent): The mouse event.
        """
        if self.ax.figure.canvas.toolbar.mode == 'zoom rect': return # does nothing when zooming
        if event.button != 1: return  # Only respond to left mouse button
        if event.inaxes != self.ax: return

        release_x = event.xdata
        left = min(self.press_x,release_x)
        width = np.abs(self.press_x - release_x)

        if np.any(self.peaks):
            if np.any(np.logical_and(self.peaks>left, self.peaks<left+width)):
                col = 'yellow'
            else:
                col = 'red'
        else:
            col= 'yellow'

        self.current_rect = Rectangle((left, self.ymin),
                                      width, self.ymax - self.ymin,
                                      facecolor=col, edgecolor='magenta', alpha=0.3)
        self.ax.add_patch(self.current_rect)
        self.rects.append(self.current_rect)

        self.press = None
        self.ax.figure.canvas.draw()

    def on_rightclick(self, event):
        """
        Handle the mouse click event for removing selected regions.

        Parameters:
        event (matplotlib.backend_bases.MouseEvent): The mouse event.
        """
        if event.inaxes != self.ax: return
        if self.ax.figure.canvas.toolbar.mode == 'zoom rect': return # does nothing when zooming
        if event.button != 3: return  # Only respond to right mouse button
        for rect in self.rects:
            contains, _ = rect.contains(event)
            if contains:
                rect.remove()
                self.rects.remove(rect)
                self.ax.figure.canvas.draw()
                break

    def get_regions(self):
        """
        Get the selected peak regions.

        Returns:
        list of tuple: List of tuples representing the start and end of each selected region in x-axis coordinates.
        """
        regions = [(rect.get_x(), rect.get_x() + rect.get_width()) for rect in self.rects]
        return regions

def select_regions(data_x, data_y, data_x2=None, data_y2=None, hkl=None,
                            xlabel='q', ylabel='Intensity',
                            title = 'Select Regions by Holding LMB, Remove by RMB',
                            max_regions=None):
    """
    Plot the 2D XRD data and allow the user to select peak regions.

    Parameters:
    data_x (array-like): The x-axis data (e.g., 2? values).
    data_y (array-like): The y-axis data (e.g., intensity values).
    max_regions (int, optional): The maximum number of regions to select. If None, an arbitrary number of regions can be selected.

    Returns:
    list of tuple: List of tuples representing the start and end of each selected region in x-axis coordinates.
    """
    plt.ion()
    fig, ax = plt.subplots(figsize=(10, 6))
    ax.plot(data_x, data_y, '-x', label='Data')
    if np.any(data_y2):
        for (xi, yi, l) in zip(data_x2, data_y2, hkl):
            ax.plot([xi,xi],[0,yi],'r')
            ax.text(xi, yi, l, va='bottom', ha='center')
        # ax.plot(data_x2, data_y2, label='Powder pattern')
        minval = min(data_y.min(), data_y2.min())
        maxval = max(data_y.max(), data_y2.max())
    else:
        minval = data_y.min()
        maxval = data_y.max()
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    ax.legend()
    selector = RegionSelector(ax, minval, maxval, 
                    max_regions=max_regions, peaks=data_x2)
    plt.show(block=True)  # Keeps the plot open until it is manually closed
    return selector#.get_regions()


####################################################
class LineSelector:
    def __init__(self, ax, data, vertical=False, func=False):
        """
        Initialize the LineSelector class.

        Parameters:
        ax (matplotlib.axes.Axes): The Axes object where the data is plotted.
        """
        self.ax = ax
        self.data = data
        self.line = None
        self.value = None
        self.vertical = vertical

        self.cidclick = ax.figure.canvas.mpl_connect('button_press_event', self.on_click)

    def on_click(self, event):
        """
        Handle the mouse click event.

        Parameters:
        event (matplotlib.backend_bases.MouseEvent): The mouse event.
        """
        if event.inaxes != self.ax: return # does nothing if the click is outside
        if self.ax.figure.canvas.toolbar.mode == 'zoom rect': return # does nothing when zooming
        if self.line is not None:
            self.line.remove()
        if self.vertical:
            self.line = self.ax.axvline(x=event.xdata, color='red', linestyle='--')
            self.value = event.xdata    
        else:
            self.line = self.ax.axhline(y=event.ydata, color='red', linestyle='--')
            self.value = event.ydata
        self.ax.figure.canvas.draw()

        n_points_selected = (self.data > self.value).sum()
        print(f'\t\tNumber of points above threshold: {n_points_selected:d}')

    def get_value(self):
        """
        Get the y-value of the last drawn horizontal line.

        Returns:
        float: The y-value of the horizontal line.
        """
        return self.value

def select_threshold(data):
    """
    Plot a 1D array and allow the user to draw a horizontal line by clicking in the window.

    Parameters:
    data (array-like): The 1D data array to plot.

    Returns:
    float: The y-value of the last drawn horizontal line.
    """
    plt.ion()
    fig, ax = plt.subplots(figsize=(10, 6))
    ax.plot(data, label='1D Data')
    ax.set_xlabel('Voxel index')
    ax.set_ylabel('SAXS signal')
    ax.set_title('Click to Draw Horizontal Line')
    selector = LineSelector(ax, data)
    # plt.legend()
    plt.show()
    plt.ioff()
    plt.show(block=True)  # Keeps the plot open until it is manually closed
    return selector.get_value()

def select_threshold_hist(data, xlabel='Intensity', ylabel='Number of Voxels', 
                          title='Click to set the lower threshold', logx=False, logy=False):
    """
    Plot a histogram and allow the user to draw a vertical line by clicking in the window.

    Parameters:
    data (array-like): The 1D data array to plot.

    Returns:
    float: The x-value of the last drawn vertical line.
    """
    plt.ion()
    fig, ax = plt.subplots(figsize=(10, 6))
    nbins = 50
    if logx:
        bins = np.geomspace(data[data>0].min(), data.max(), nbins+1)
        # bins = np.logspace(np.log10(data[data>0].min()),np.log10(data.max()), nbins)
    else:
        bins=nbins

    ax.hist(data, bins=bins, label='1D Data')
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    if logx:
        ax.set_xscale('log')
    if logy:
        ax.set_yscale('log')
    selector = LineSelector(ax, data, vertical=True)
    # plt.legend()
    plt.show()
    plt.ioff()
    plt.show(block=True)  # Keeps the plot open until it is manually closed
    plt.ion()
    return selector.get_value()

def mask_voxels( Kappa, Omega, ty, tz, shifty, shiftz, x_p, tomogram, sinogram ):
    print('Identify your sample by drawing a rectangle over it')
    # find the images at angles to check sample geometry
    to_try = np.array([[0,0],[0,np.pi/2]])
    rgl = []
    for b in to_try:
        A = np.column_stack((Kappa, Omega % 2*np.pi))
        # check which of the projections is closest to the angles defined in totry
        distances = np.linalg.norm(A - b, axis=1)
        g_match = np.argmin(distances)
        # plot the projection in textom scale
        Y,Z = np.meshgrid( ty+shifty[g_match], tz+shiftz[g_match] )
        rgl.append( draw_rectangle_on_image( 
            sinogram[:,:,g_match].T, xy=(Y,Z),
            title=f'Projection No {g_match}, tilt = {A[g_match,0]*180/np.pi} \u00b0, rot = {A[g_match,1]*180/np.pi} \u00b0' 
        ))

    zmin = np.min( [rgl[0].start_point[1], rgl[0].end_point[1],
                    rgl[1].start_point[1], rgl[1].end_point[1],])
    zmax = np.max( [rgl[0].start_point[1], rgl[0].end_point[1],
                    rgl[1].start_point[1], rgl[1].end_point[1],])
    ymin = min( rgl[0].start_point[0], rgl[0].end_point[0] )
    ymax = max( rgl[0].start_point[0], rgl[0].end_point[0] )
    xmin = min( rgl[1].start_point[0], rgl[1].end_point[0] )
    xmax = max( rgl[1].start_point[0], rgl[1].end_point[0] )
    vmask_0 = np.logical_and.reduce((
            x_p[:,0] > xmin, x_p[:,0] < xmax,
            x_p[:,1] > ymin, x_p[:,1] < ymax,
            x_p[:,2] > zmin, x_p[:,2] < zmax,
    ))

    happy = 'n'
    while happy != 'y':
        print('\tMasking empty voxels. Choose lower threshold in figure.')
        tomo_flat = tomogram.flatten()
        # # draw all data in a sausage and choose a threshold directly
        # self.cutoff_low = msk.select_threshold( tomo_flat[vmask_0] )
        # draw a horizontal cutoff in a histogram
        cutoff_low = select_threshold_hist( tomo_flat[vmask_0] )

        # apply the cutoff and region exclusion to the mask
        vmask = np.logical_and.reduce((
            tomo_flat > cutoff_low,
            vmask_0))

        # # erosion dilation of noise
        # structure_size = 2
        # structuring_element = ndimage.generate_binary_structure(3, 1)
        # structuring_element = ndimage.iterate_structure(
        #     structuring_element, structure_size)
        # vmask = ndimage.binary_opening(vmask, structure=structuring_element)

        # smooth surface with a filter (and grow slightly)
        structure = ndimage.generate_binary_structure(3, 1)
        filled = ndimage.binary_fill_holes(vmask.reshape(tomogram.shape), structure=structure)
        smoothed = ndimage.gaussian_filter(filled.astype(float), sigma=1) > 0.4
        # keep largest component
        labeled, n = ndimage.label(smoothed)
        sizes = ndimage.sum(smoothed, labeled, range(1, n + 1))
        largest = labeled == (np.argmax(sizes) + 1)
        vmask = largest.flatten()
        print(f'\tSmoothed surface. New number of voxels: {vmask.sum()}')

        # print('\tMasking empty voxels. Choose regions to exclude in figure.')
        # sphere_mask = msk.sphere_exclusion(x_p[vmask])
        # vmask[vmask] = sphere_mask
        
        mask_voxels = np.where( vmask )[0]
        check_tomogram(tomogram, mask_voxels)
        happy = input('\thappy? (y/n) ')
    
    return mask_voxels

def check_tomogram( tomogram, mask ):
    """Plots the surface of a masked 3D object

    Parameters
    ----------
    tomogram : 4D ndarray, float
        data shape (nx,ny,nz,1)
    mask : 1D ndarray, int
        list of not masked voxels
    """
    masked_array = np.zeros_like(tomogram.flatten())
    masked_array[mask] = True
    masked_array = masked_array.reshape(tomogram.shape[:-1])
    
    verts, faces, _, _ = marching_cubes(masked_array, level=0.5, step_size=1)
    
    fig = plt.figure()#figsize=(10, 10))
    ax = fig.add_subplot(111, projection='3d')
    
    ax.plot_trisurf(verts[:, 0], verts[:, 1], verts[:, 2], triangles=faces, cmap='viridis', lw=0.2, alpha=0.5)

    # Add labels for clarity
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")
    ax.set_title("Surface of the masked object (arbitrary colors). Verify and close figure")

    set_axes_equal(ax)
    plt.show(block=True)

def mask_detector( image, mask ):
    """Creates a mask on a 2D image by user-input

    Left-click on pixels to exclude, right-click to include
    Close window to confirm

    Parameters
    ----------
    image : 2D ndarray, float
        example image
    mask : 2D ndarray, bool
        a priori guess for the mask

    Returns
    -------
    2D ndarray, bool
    """
    maxval = image.max()

    fig, ax = plt.subplots( figsize = (10,10))
    ax.imshow(image, norm=colors.LogNorm(vmin=maxval*1e-2, vmax=maxval) )
    # Overlay the mask in red
    mask_color = np.zeros((*mask.shape, 4))  # Create an RGBA array
    mask_color[~mask] = [1, 0, 0, 0.5]  # Red with 50% transparency for True values
    ax.imshow( mask_color ) # plot the mask on top
    ax.set_title('Click pixels or draw lines. LMB mask, RMB unmask')

    def on_click(event):
        global start_point
        if event.inaxes != ax:
            return
        start_point = (event.xdata, event.ydata)

    def on_release(event):
        global start_point
        if event.inaxes != ax or start_point is None:
            return
        # calculate the line in pixels
        end_point = (event.xdata, event.ydata)
        distance = np.sqrt(
            (end_point[0] - start_point[0])**2 +\
            (end_point[1] - start_point[1])**2)
        if distance > 0.5:
            steps = np.arange(0, distance, step=0.5)
            points = np.unique(
                ( 
                np.round(start_point[0] + steps*(end_point[0]-start_point[0])/distance).astype(int),
                np.round(start_point[1] + steps*(end_point[1]-start_point[1])/distance).astype(int) ),
                axis=1, 
            ).T
        else:
            points = np.array([np.round(end_point)],dtype=int)

        if event.button == 1:  # Left mouse button: mask
            for p in points:
                mask[ p[1], p[0] ] = 0

        elif event.button == 3:  # Right mouse button: unmask
            for p in points:
                mask[ p[1], p[0] ] = 1

        # replot
        plt.gca().clear() 
        ax.set_title('Click pixels or draw lines. LMB mask, RMB unmask')
        ax.imshow(image, norm=colors.LogNorm(vmin=maxval*1e-2, vmax=maxval) )
        # Overlay the mask in red
        mask_color = np.zeros((*mask.shape, 4))  # Create an RGBA array
        mask_color[~mask] = [1, 0, 0, 0.5]  # Red with 50% transparency for True values
        ax.imshow( mask_color ) # plot the mask on top
        fig.canvas.draw_idle()

    cid_click = fig.canvas.mpl_connect('button_press_event', on_click)
    cid_release = fig.canvas.mpl_connect('button_release_event', on_release)

    plt.show(block=True)

    plt.figure()
    plt.imshow(mask)
    plt.title('Dark regions are masked')
    plt.show()

    return mask#.flatten()

####################################################
class RectangleDrawer:
    def __init__(self, ax, image):
        self.ax = ax
        self.image = image
        self.rect = None
        self.press = None
        self.start_point = None
        self.end_point = None
        self.cid_press = ax.figure.canvas.mpl_connect('button_press_event', self.on_press)
        self.cid_release = ax.figure.canvas.mpl_connect('button_release_event', self.on_release)
        self.cid_motion = ax.figure.canvas.mpl_connect('motion_notify_event', self.on_motion)

    def on_press(self, event):
        if event.inaxes != self.ax:
            return
        if self.rect is not None:
            self.rect.remove()
            self.rect = None
            self.ax.figure.canvas.draw()
        self.start_point = (event.xdata, event.ydata)
        self.rect = plt.Rectangle(self.start_point, 0, 0, fill=False, edgecolor='red')
        self.ax.add_patch(self.rect)
        self.press = event.xdata, event.ydata

    def on_motion(self, event):
        if self.press is None:
            return
        if event.inaxes != self.ax:
            return
        x0, y0 = self.press
        x1, y1 = event.xdata, event.ydata
        self.rect.set_width(x1 - x0)
        self.rect.set_height(y1 - y0)
        self.ax.figure.canvas.draw()

    def on_release(self, event):
        if self.press is None:
            return
        if event.inaxes != self.ax:
            return
        self.press = None
        self.end_point = (event.xdata, event.ydata)
        self.ax.figure.canvas.draw()

    def disconnect(self):
        self.ax.figure.canvas.mpl_disconnect(self.cid_press)
        self.ax.figure.canvas.mpl_disconnect(self.cid_release)
        self.ax.figure.canvas.mpl_disconnect(self.cid_motion)

def draw_rectangle_on_image(image, xy = None, title=''):
    fig, ax = plt.subplots()

    if xy:
        ax.pcolormesh(xy[0], xy[1], image, cmap='jet', 
                    #   norm=colors.LogNorm()
                      )
    else:
        ax.imshow(image, cmap='jet', 
                  #norm=colors.LogNorm()
                  )
    ax.set_title(title)

    drawer = RectangleDrawer(ax, image)
    plt.show(block=True)
    return drawer

# ####################################################
# def sphere_exclusion( data ):
#     # Initialize the plot
#     fig = plt.figure(figsize=(10, 6))
#     ax = fig.add_subplot(111, projection='3d')
#     ax.set_box_aspect([1, 1, 1])
#     ax.set_title("3D Scatter with Interactive Spheres")
#     ax.set_xlabel('x')
#     ax.set_ylabel('y')
#     ax.set_zlabel('z')

#     points = filter_grid_points(data)

#     # Initial scatter plot data
#     x = points[:,0] # np.random.rand(10)
#     y = points[:,1] # np.random.rand(10)
#     z = points[:,2] # np.random.rand(10)
#     colors = np.full(len(points), 'blue')  # Initial color for all points
#     scatter = ax.scatter(x, y, z, color=colors)

#     # Lists to store sphere data and plot objects
#     spheres = []
#     sphere_patches = []

#     # Lists to store sphere data (center and radius) and plot objects
#     spheres = []
#     sphere_patches = []

#     # Function to check if points are inside any sphere
#     def points_within_spheres(points, spheres):
#         bool_array = np.zeros(len(points), dtype=bool)
#         for (x_center, y_center, z_center, radius) in spheres:
#             distances = np.sqrt((points[:, 0] - x_center) ** 2 +
#                                 (points[:, 1] - y_center) ** 2 +
#                                 (points[:, 2] - z_center) ** 2)
#             bool_array |= distances <= radius
#         return bool_array
    
#     # Function to update scatter plot colors based on spheres
#     def update_point_colors():
#         inside_spheres = points_within_spheres(points, spheres)
#         colors = np.where(inside_spheres, 'red', 'blue')
#         scatter._facecolor3d = colors
#         scatter._edgecolor3d = colors
#         fig.canvas.draw_idle()

#     # Function to add a sphere
#     def add_sphere(event):
#         try:
#             x_center = float(text_box_x.text)
#             y_center = float(text_box_y.text)
#             z_center = float(text_box_z.text)
#             radius = float(text_box_radius.text)

#             # Generate sphere points for plotting
#             u = np.linspace(0, 2 * np.pi, 30)
#             v = np.linspace(0, np.pi, 30)
#             x_sphere = x_center + radius * np.outer(np.cos(u), np.sin(v))
#             y_sphere = y_center + radius * np.outer(np.sin(u), np.sin(v))
#             z_sphere = z_center + radius * np.outer(np.ones(np.size(u)), np.cos(v))

#             # Plot the sphere and add it to the lists
#             sphere = ax.plot_surface(x_sphere, y_sphere, z_sphere, color='r', alpha=0.3)
#             spheres.append((x_center, y_center, z_center, radius))
#             sphere_patches.append(sphere)

#             # Update point colors based on new spheres
#             update_point_colors()
#         except ValueError:
#             print("Please enter valid numerical values for center and radius.")

#     # Function to delete the last sphere added
#     def delete_sphere(event):
#         if sphere_patches:
#             sphere = sphere_patches.pop()
#             sphere.remove()
#             spheres.pop()
#             # Update point colors based on remaining spheres
#             update_point_colors()
#         else:
#             print("No spheres to delete.")

#     # Event handler for window close to return the boolean array
#     def on_close(event):
#         fig.canvas.stop_event_loop()  # Stop event loop if run interactively

#     fig.canvas.mpl_connect('close_event', on_close)

#     # Create TextBox and Button widgets on the right side of the plot
#     plt.subplots_adjust(left=0.05, right=0.75)

#     axbox_x = plt.axes([0.8, 0.75, 0.15, 0.05])
#     text_box_x = TextBox(axbox_x, 'X Center', initial="0")

#     axbox_y = plt.axes([0.8, 0.65, 0.15, 0.05])
#     text_box_y = TextBox(axbox_y, 'Y Center', initial="0")

#     axbox_z = plt.axes([0.8, 0.55, 0.15, 0.05])
#     text_box_z = TextBox(axbox_z, 'Z Center', initial="0")

#     axbox_radius = plt.axes([0.8, 0.45, 0.15, 0.05])
#     text_box_radius = TextBox(axbox_radius, 'Radius', initial="1")

#     add_button_ax = plt.axes([0.8, 0.35, 0.15, 0.05])
#     add_button = Button(add_button_ax, 'Add Sphere')
#     add_button.on_clicked(add_sphere)

#     delete_button_ax = plt.axes([0.8, 0.25, 0.15, 0.05])
#     delete_button = Button(delete_button_ax, 'Delete Sphere')
#     delete_button.on_clicked(delete_sphere)

#     set_axes_equal(ax)

#     # Display the plot and wait for it to close
#     plt.show(block=True)

#     # Calculate and return the boolean array indicating which points are in any sphere
#     cheese = ~points_within_spheres(data, spheres)
#     return cheese

def set_axes_equal(ax):
    """
    Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc.

    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().
    """

    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence I call half the max range the plot radius.
    plot_radius = 0.5*max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])



### base coordinate system: ###
beam_direction = (1,0,0) # p in mumott
transverse_horizontal = (0,1,0) # j in mumott
transverse_vertical = (0,0,1) # k in mumott
################################ don't change

# detector geometry: 
flip_detector_ud=False 
flip_detector_lr=False
detector_direction_origin = (0,-1,0) # this conforms to standard pyFAI output
detector_direction_positive_90 = (0,0,-1) # this conforms to standard pyFAI output

# sample movements:
inner_axis = (0,0,1) # inner rotation axis
outer_axis = (0,1,0) # outer rotation axis
scan_mode = 'line' #'column' # 'line_snake' # 'column_snake' # 'controt'
flip_fov=False # flip fast and slow axis argument (for the case it was defined the reverse way in the integrated data files)

# For calculating projectors:
Dbeam = 0.3 # beam size in um (FWHM)
Dstep = 0.5 # scanning step size in um
